from PymoNNto import *
from PymoNNto.Exploration.Visualization.Reconstruct_Analyze_Label.Reconstruct_Analyze_Label import *

def compute_temporal_reconstruction(network, single_neuron_group=None, single_neuron_id=None, recon_group_tag='recon'):
    RALN = Reconstruct_Analyze_Label_Network(network)
    RALN.zero_recon()
    if single_neuron_id is not None and single_neuron_group is not None:
        single_neuron_group.recon[single_neuron_id] = 1
    else:
        for ng in network.NeuronGroups:
            ng.recon = ng.output.copy()
    RALN.propagation('W', 10, 'backward', 'forget', 'all', temporal_recon_groups=network[recon_group_tag], exponent=4, normalize=True, filter_weakest_percent=40.0)  # forget

    for ng in network[recon_group_tag]:
        baseline = ng.Input_Weights.transpose().dot(ng.temporal_recon[-1])
        baseline = baseline / np.sum(baseline)

        if single_neuron_id is not None and single_neuron_group is not None and ng == single_neuron_group:  # clicked group?
            temp = ng.get_neuron_vec()
            temp[single_neuron_id] = 1.0
            ng.temporal_recon.insert(0, temp)
        else:
            ng.temporal_recon.append(ng.get_neuron_vec())


        res = []

        for r in ng.temporal_recon:
            char_vec = ng.Input_Weights.transpose().dot(r)
            s = np.sum(char_vec)
            if s > 0:
                char_vec = char_vec / s
                res.insert(0, char_vec - baseline)
                #text = generator.index_to_char(np.argmax(char_vec)) + text
            else:
                res.insert(0, np.zeros(ng.Input_Weights.shape[1]))
                #text = '#' + text

        #text = list(text)
        #start = text[0]
        #for i, c in enumerate(text):
        #    if text[i] == start:
        #        text[i] = '#'
        #    else:
        #        break
        #text = ''.join(text)

        return res

def generate_text_from_recon_mat(recon_mat, generator):
    text = ''
    for char_vec in recon_mat:
        s = np.max(char_vec)-np.min(char_vec)#np.sum(char_vec)
        if s>0:
            text = text + generator.index_to_char(np.argmax(char_vec))
        else:
            text = text + '#'

    text = list(text)
    start = text[0]
    for i, c in enumerate(text):
        if text[i] == start:
            text[i] = '#'
        else:
            break
    text = ''.join(text)
    return text.replace(' ','_')


from PymoNNto.Exploration.UI_Base import *
from PymoNNto.NetworkBehaviour.Recorder.Recorder import *
from PymoNNto.Exploration.Network_UI.Neuron_Classification_Colorizer import *

class Network_UI(UI_Base):

    def __init__(self, network, modules=[], label='SORN UI', group_tags=[], transmitters=[], storage_manager=None, group_display_count=None, reduced_layout=False):

        network.simulate_iteration()

        network.clear_recorder()

        self.render_every_x_frames = 1

        #network.simulate_iteration()
        #self.recording = False
        #if self.recording:
        #    label += ' rec.'

        for ng in network.NeuronGroups:
            if not hasattr(ng, 'color'):
                ng.color = (0, 0, 255, 255)
            ng.add_analysis_module(Neuron_Classification_Colorizer())

        if group_tags==[]:
            for ng in network.NeuronGroups:
                if ng.tags[0] not in group_tags:
                    group_tags.append(ng.tags[0])

        if transmitters==[]:
            for sg in network.SynapseGroups:
                if sg.tags[0] not in transmitters:
                    transmitters.append(sg.tags[0])

        self.reduced_layout=reduced_layout

        #for group in network[inh_group_name]:
        #    network.add_behaviours_to_neuron_group({10000: Recorder(['np.mean(n.output)',
        #                           'np.mean(n.TH)',
        #                           'n.TH',
        #                           'n.excitation',
        #                           'n.inhibition',
        #                           'n.input_act',
        #                           'n.refractory_counter',
        #                           '[np.sum(s.slow_add) for s in n.afferent_synapses.get("All")]',
        #                           '[np.sum(s.fast_add) for s in n.afferent_synapses.get("All")]'], tag='UI_rec')}, group)

        super().__init__(network, label=label)

        self.group_display_count = group_display_count

        #self.exc_group_name = exc_group_name
        #self.inh_group_name = inh_group_name
        self.group_tags = group_tags
        self.transmitters=transmitters
        self.pause = False
        self.update_without_state_change = False
        self.storage_manager = storage_manager

        self.neuron_visible_groups = []

        #self.neuron_select_id = 0
        self._neuron_select_group = network[group_tags[0], 0]
        self._neuron_select_mask = self._neuron_select_group.get_neuron_vec().astype(np.bool)#np.array([0])
        self._neuron_select_mask[0] = True

        self.neuron_select_x = 0
        self.neuron_select_y = 0


        #self.ts_group = 0
        #self.x_steps = 500
        #self.group_sliders = []
        self.neuron_select_color = (0, 255, 0, 255)

        self.modules = modules

        if type(self.modules) is dict:
            self.modules = self.modules.values()

        for module in self.modules:
            print('Initialize:', module)
            module.initialize(self)

        for group_tag in group_tags:
            for group in network[group_tag]:

                group._rec_dict = {}

                #rec = Recorder([], tag='UI_rec')
                #network.add_behaviours_to_neuron_group({10000: rec}, group)

                for module in self.modules:
                    module.add_recorder_variables(group, self)

        self.init_recoders()

        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.on_timer)
        self.timer.start(40)


    def select_neuron_class(self, group, class_id, add_to_select_group=False):
        return self.select_neurons(group, group.classification == class_id, add_to_select_group)

    def select_neuron(self, group, id, add_to_select_group=False):
        return self.select_neurons(group, id, add_to_select_group)

    def select_neurons(self, group, mask, add_to_select_group=False):
        group_changed = group != self._neuron_select_group
        self._neuron_select_group = group
        if group_changed or not add_to_select_group:
            self._neuron_select_mask = self._neuron_select_group.get_neuron_vec().astype(np.bool)
        self._neuron_select_mask[mask] = True

        print(np.where(self._neuron_select_mask>0)[0])

        for module in self.modules:
            module.on_selected_neuron_changed(self)
        self.static_update_func()


    def selected_neuron_tag(self):
        return self._neuron_select_group.tags[0]

    def selected_neuron_group(self):
        return self._neuron_select_group

    def selected_neuron_mask(self):
        return self._neuron_select_mask

    def selected_neuron_id(self):
        return self._neuron_select_mask.argmax()#first "True" value in array



    def add_recording_variable(self, group, var, timesteps):

        try:
            n = group  # for eval
            eval(var) #produce error when not evaluable

            old_ts=0
            if var in group._rec_dict:
                old_ts=group._rec_dict[var]

            group._rec_dict[var] = max(timesteps,old_ts)
            #recorder.add_varable('n.output')
            return True
        except:
            return False

    def init_recoders(self):
        for group_tag in self.group_tags:
            for group in self.network[group_tag]:

                rec_time_dict={}
                for variable in group._rec_dict:
                    rec_length=group._rec_dict[variable]
                    if rec_length not in rec_time_dict:
                        rec_time_dict[rec_length]=[]
                    rec_time_dict[rec_length].append(variable)

                for rec_length in rec_time_dict:
                    rec = Recorder(rec_time_dict[rec_length] + ['n.iteration'], tag='UI_rec,rec_' + str(rec_length), max_length=rec_length)
                    self.network.add_behaviours_to_object({10000+rec_length: rec}, group)

    #def rec(self, neuron_group, rec_length=-1):
    #    return neuron_group[self.rec_tag(rec_length),0]

    #def rec_tag(self, rec_length=-1):
    #    if rec_length==-1:
    #        rec_length = self.default_rec_recording_length
    #    return 'rec_'+str(rec_length)

    def static_update_func(self, event=None):
        if self.pause:
            self.update_without_state_change=True

    def get_selected_neuron_subgroup(self):
        syn_sgs = self.get_selected_synapses()
        if len(syn_sgs) > 0:
            return syn_sgs[0]
        else:
            return None

    def get_selected_synapses(self):
        result = []
        group = self.selected_neuron_group()
        synapse_groups = group.afferent_synapses['All']
        for i, s in enumerate(synapse_groups):
            if (type(s.dst.mask) == np.ndarray and s.dst.mask[self.selected_neuron_id()]) or (type(s.dst.mask) is bool and s.dst.mask == True):
                result.append(synapse_groups[i].dst)
        return result

    def on_timer(self):

        if not self.pause or self.step or self.update_without_state_change:
            self.step = False
            if not self.update_without_state_change:
                for i in range(self.render_every_x_frames):
                    self.network.simulate_iteration()

            self.it = self.network.iteration

            for module in self.modules:
                module.update(self)

            #for rec in self.network['UI_rec']:
            #    rec.cut_length(self.default_recorder_length)

            self.update_without_state_change = False

    def on_tab_change(self, i):
        self.static_update_func()

def get_color(type_index, layer=1):
    dim_value = max(layer * 0.9, 1.0)

    if type_index == 0:
        return (0.0, 0.0, 255.0 / dim_value, 255.0)
    if type_index == 1:
        return (255.0 / dim_value, 0.0, 0.0, 255.0)
    if type_index == 2:
        return (255.0 / dim_value, 150.0 / dim_value, 0.0, 255.0)
    if type_index == 3:
        return (255.0 / dim_value, 80.0 / dim_value, 0.0, 255.0)
    if type_index == 4:
        return (255.0 / dim_value, 0.0 , 150.0/ dim_value, 255.0)


class Analytics_Results_Select_ComboBox(QComboBox):
    popupAboutToBeShown = QtCore.pyqtSignal()

    def __init__(self, main_object, tag='classifier', first_entry=None, select_last=True, on_update_func=None):
        super().__init__()

        self.on_update_func = on_update_func

        self.first_entry = first_entry
        self.tag = tag

        self.current_results = {}
        self.current_modules = {}

        self.main_object = None
        self.change_main_object(main_object)

        if select_last and self.count() > 0:
            self.setCurrentIndex(self.count()-1)



    def change_main_object(self, main_object):
        if self.main_object is not None:
            self.remove_update_notifier()
        self.main_object = main_object
        self.update()
        self.set_update_notifier()

    def set_update_notifier(self):
        if isinstance(self.main_object, AnalysisModule):
            self.main_object.set_update_notifier(self.update)
        else:
            for obj in self.main_object[self.tag]:
                if isinstance(obj, AnalysisModule):
                    obj.set_update_notifier(self.update)

    def remove_update_notifier(self):
        if isinstance(self.main_object, AnalysisModule):
            self.main_object.remove_update_notifier(self.update)
        else:
            for obj in self.main_object[self.tag]:
                if isinstance(obj, AnalysisModule):
                    obj.remove_update_notifier(self.update)

    def get_selected_key(self):
        return self.currentText()

    def get_selected_module(self):
        key = self.get_selected_key()
        if key in self.current_modules:
            return self.current_modules[key]

    def get_selected_result(self):
        key = self.get_selected_key()
        if key in self.current_results:
            return self.current_results[key]

    def update(self, key=None):
        if key is None:
            current = self.currentText()
        else:
            current = key

        self.clear()

        if self.first_entry is not None:
            self.addItem(self.first_entry)

        if isinstance(self.main_object, AnalysisModule):
            self.current_results = self.main_object.get_results()
            for k in self.current_results:
                self.addItem(k)
            self.current_modules = {k: self.main_object for k in self.current_results}
        else:
            self.current_results, self.current_modules = self.main_object.get_all_analysis_module_results(self.tag, True)
            for k in self.current_results:
                self.addItem(k)

        for i in range(self.count()):
            if self.itemText(i) == current:
                self.setCurrentIndex(i)

        if self.on_update_func is not None:
            self.on_update_func(self.main_object, self.get_selected_module(), self.get_selected_key(), self.get_selected_result())



########################################################### Exception handling


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)

sys.excepthook = except_hook





from PymoNNto.Exploration.Network_UI.Network_UI import *
from PymoNNto.Exploration.Network_UI.TabBase import *

class sidebar_neuron_grid_submodule(TabBase):

    def __init__(self, add_color_dict):
        self.add_color_dict = add_color_dict

        self.compiled = {}
        for param in self.add_color_dict:
            self.compiled[param] = compile('n.'+param, '<string>', 'eval')

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def on_selected_neuron_changed(self, Network_UI):
        tag = self.image_item.neuron_group_tag
        self.color_select_box.change_main_object(Network_UI.network[tag, 0])

    def get_clicked_neuron_id(self, Network_UI, event):
        group_tag = event.currentItem.neuron_group_tag
        w = Network_UI.network[group_tag, 0].width
        h = Network_UI.network[group_tag, 0].height
        d = Network_UI.network[group_tag, 0].depth
        y_temp = int(
            (h * d - 1) - np.trunc(event.pos().y()))  # np.clip(int((h - 1) - np.trunc(event.pos().y())), 0, h - 1)

        neuron_select_x = np.clip(int(np.trunc(event.pos().x())), 0, w - 1)
        neuron_select_y = np.clip(int(y_temp - np.trunc(y_temp / h)), 0, h - 1)
        neuron_select_z = np.clip(int(np.trunc(y_temp / h)), 0, d - 1)

        h_abs = np.clip(int(y_temp), 0, h * d - 1)
        return (h_abs) * w + neuron_select_x

    def initialize(self, Network_UI, index):


        #just store for mcde event bug ('QGraphicsSceneMouseEvent' object has no attribute 'currentItem')
        self.last_id = 0
        self.last_group = None

        def mce(event):
            self.last_id = self.get_clicked_neuron_id(Network_UI, event)
            self.last_group = Network_UI.network[event.currentItem.neuron_group_tag, 0]
            Network_UI.select_neuron(self.last_group, self.last_id, add_to_select_group=Network_UI.control_key_down)

        def mdce(event):
            if self.last_group.classification is not None:
                selected_class = self.last_group.classification[self.last_id]
                Network_UI.select_neuron_class(self.last_group, selected_class, add_to_select_group=Network_UI.control_key_down)


        group_select_box = QComboBox()
        self.color_select_box = Analytics_Results_Select_ComboBox(Network_UI.network[Network_UI.group_tags[index], 0] ,'classifier', first_entry='group color')

        '''
        def update_list():
            self.color_select_box.clear()
            self.color_select_box.addItem('color')

            group = Network_UI.network[Network_UI.group_tags[index], 0]
            for k in group.get_all_analysis_module_results('classifier'):
                self.color_select_box.addItem(k)

        self.color_select_box.popupAboutToBeShown.connect(update_list)
        '''

        #update_list()
        #if self.color_select_box.count()>1:
        #    self.color_select_box.setCurrentIndex(1)#select default if exist


        Network_UI.Add_Sidebar_Element([group_select_box, self.color_select_box])

        self.image_item = Network_UI.Add_Image_Item(False, True, tooltip_message='white: active neurons\r\ndarker color: primary input neurons\r\ngreen: selected neuron')

        self.image_item.neuron_group_tag = Network_UI.group_tags[index]
        Network_UI.neuron_visible_groups.append(Network_UI.group_tags[index])
        self.image_item.mouseClickEvent = mce

        self.image_item.mouseDoubleClickEvent = mdce



        def group_changed(select_index):
            tag = Network_UI.group_tags[select_index]
            Network_UI.select_neuron(Network_UI.network[tag, 0], 0)
            self.image_item.neuron_group_tag = tag
            Network_UI.neuron_visible_groups[index] = tag

        group_select_box.addItems(Network_UI.group_tags)
        group_select_box.setCurrentIndex(index)
        group_select_box.currentIndexChanged.connect(group_changed)



    def update(self, Network_UI):

        #for group_tag in Network_UI.group_tags:
        group_tag = self.image_item.neuron_group_tag
        if len(Network_UI.network[group_tag]) > 0:

            group = Network_UI.network[group_tag, 0]
            n = group #for eval operations

            #base color and classification
            #key=self.color_select_box.currentText()
            #classifications = group.get_all_analysis_module_results('classifier')
            #if key in classifications:
            #    classification = classifications[key]
            #else:
            #    classification = None
            group.classification = self.color_select_box.get_selected_result()
            image = np.array(group['Neuron_Classification_Colorizer', 0].get_color_list(group.classification, format='[r,g,b]'))

            #selected
            if Network_UI.selected_neuron_group().tags[0] == group_tag:
                for i in range(3):
                    image[Network_UI.selected_neuron_mask(), i] = Network_UI.neuron_select_color[i]

            #parameter
            for param, color in self.add_color_dict.items():
                try:
                    data=eval(self.compiled[param])

                    if (type(data) == np.ndarray and data.dtype == np.dtype('bool')) or (type(data) == bool):
                        for i in range(3):
                            image[data, i] += color[i]
                    else:
                        for i in range(3):
                            image[:, i] += color[i]*data
                except:
                    pass



            image=np.reshape(image, (group.height*group.depth, group.width, 3))#group.depth
            self.image_item.setImage(np.rot90(image, 3), levels=(0, 255))

class UI_sidebar_neuron_grid_module():

    def __init__(self, group_display_count=1, add_color_dict={'output': (255, 255, 255)}):
        self.group_display_count = group_display_count
        if type(add_color_dict) is str:
            add_color_dict={add_color_dict:(255,255,255)}
        self.add_color_dict = add_color_dict

    def add_recorder_variables(self, neuron_group, Network_UI):
        for module in self.sub_modules:
            module.add_recorder_variables(neuron_group, Network_UI)

    def on_selected_neuron_changed(self, NetworkUI):
        for module in self.sub_modules:
            module.on_selected_neuron_changed(NetworkUI)

    def initialize(self, Network_UI):

        if Network_UI.group_display_count is not None:
            self.group_display_count=Network_UI.group_display_count

        self.sub_modules = []
        for i in range(self.group_display_count):
            self.sub_modules.append(sidebar_neuron_grid_submodule(self.add_color_dict))
            self.sub_modules[-1].initialize(Network_UI, np.minimum(i, len(Network_UI.group_tags)))

    def update(self, Network_UI):
        for module in self.sub_modules:
            module.update(Network_UI)


from PymoNNto import *
import sys
import matplotlib.pyplot as plt
from PyQt5 import QtCore, QtGui, QtWidgets, uic
from PyQt5 import QtCore
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import pyqtgraph as pg
import numpy as np

from PyQt5 import QtCore, QtGui, QtWidgets, uic
from PyQt5.QtCore import Qt
from PIL import Image

#from X_Experimental.Functions import *
#from Testing.SORN.SORN_Helper import *


class TabBase:

    def __init__(self, title='Title'):
        self.title = title

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        return

    def update(self, Network_UI):
        return

    def on_selected_neuron_changed(self, Network_UI):
        return



from PymoNNto.Exploration.Network_UI.TabBase import *

class DrawItem(pg.GraphicsObject):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.picture = QtGui.QPicture()
        #self.update()

    def get_rnd_color(self):
        return (np.random.rand()*255.0,np.random.rand()*255.0,np.random.rand()*255.0,255.0)

    def get_rect(self, ng, x_shift=0, y_shift=0):
        minx = np.min(ng.x)
        miny = np.min(ng.y)
        maxx = np.max(ng.x)
        maxy = np.max(ng.y)
        return QtCore.QRectF(minx + x_shift, miny+y_shift, maxx - minx, maxy - miny)

    def draw_input_output_block(self, painter, sg, fill=False, y_shift=0):
        painter.setPen(pg.mkPen(color=(0, 0, 0, 255)))
        if fill:
            painter.setBrush(pg.mkBrush(color=sg.src.color))
        else:
            painter.setBrush(pg.mkBrush(color=(0, 0, 0, 0)))
        painter.drawRect(self.get_rect(sg.src, -30, y_shift))

        painter.setPen(pg.mkPen(color=(0, 0, 0, 255)))
        if fill:
            painter.setBrush(pg.mkBrush(color=sg.dst.color))
        else:
            painter.setBrush(pg.mkBrush(color=(0, 0, 0, 0)))
        sg.dst_position_rect=self.get_rect(sg.dst, 30.0)
        painter.drawRect(sg.dst_position_rect)


    def update_pic(self, sub_sgs, group):

        blocks={}

        for sg in sub_sgs:
            if not sg.src.BaseNeuronGroup in blocks:
                blocks[sg.src.BaseNeuronGroup]=[]
            blocks[sg.src.BaseNeuronGroup].append(sg)

        self.sub_sgs=sub_sgs

        self.picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.picture)

        painter.setPen(pg.mkPen(color=(0, 0, 0, 100)))
        painter.setBrush(pg.mkBrush(color=(0, 0, 0, 0)))

        for x,y in zip(group.x, group.y):
            painter.drawEllipse(QtCore.QPointF(x+30, y), 0.4, 0.4)

        if len(sub_sgs)>0:
            for x,y in zip(sub_sgs[0].src.BaseNeuronGroup.x,sub_sgs[0].src.BaseNeuronGroup.y):
                painter.drawEllipse(QtCore.QPointF(x-30, y), 0.4, 0.4)

        for i, key in enumerate(blocks):
            for sg in blocks[key]:
                if hasattr(sg, 'selected') and sg.selected:
                    self.draw_input_output_block(painter, sg, True, 30*i)

        for i, key in enumerate(blocks):
            for sg in blocks[key]:
                self.draw_input_output_block(painter, sg, False, 30*i)

        painter.drawPolygon(QtCore.QPoint(0,1),
                            QtCore.QPoint(0,2),
                            QtCore.QPoint(3,0),
                            QtCore.QPoint(0,-2),
                            QtCore.QPoint(0,-1),
                            QtCore.QPoint(-4,-1),
                            QtCore.QPoint(-4,1))

        painter.end()
        self.prepareGeometryChange()
        self.informViewBoundsChanged()
        #self.update()

    def paint(self, painter, option, widget=None):
        painter.drawPicture(0, 0, self.picture)

    def boundingRect(self):
        br=self.picture.boundingRect()
        br.setTop(min(br.top(), br.left()))
        br.setLeft(min(br.top(), br.left()))
        br.setRight(max(br.right(), br.bottom()))
        br.setBottom(max(br.right(), br.bottom()))
        return QtCore.QRectF(br)


class partition_tab(TabBase):

    def __init__(self, title='Partition'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.partition_tab = Network_UI.Next_Tab(self.title)

        def on_click(event):
            for sg in event.currentItem.sub_sgs:
                sg.selected=sg.dst_position_rect.contains(event.pos().x(), event.pos().y())

        self.draw_items = {}
        self.plots = {}
        for transmitter in Network_UI.transmitters:
            self.plots[transmitter] = Network_UI.Add_plot(tooltip_message='Click on a block of the right group(selected group) to show the possible inputs to this block\r\nLeft: Source NeuronGroup\r\nRight: Destination NeuronGroup\r\n(Partitioning is used for speeding up synapse operations)')
            self.draw_items[transmitter] = DrawItem()
            self.draw_items[transmitter].mouseClickEvent = on_click
            self.plots[transmitter].addItem(self.draw_items[transmitter])


    def update(self, Network_UI):
        if self.partition_tab.isVisible():
            group=Network_UI.selected_neuron_group()
            for transmitter in Network_UI.transmitters:
                self.draw_items[transmitter].update_pic(group.afferent_synapses[transmitter],group)
                self.plots[transmitter].update()

            #self.plot.refresh()


from PymoNNto.Exploration.Network_UI.Network_UI import *
from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Analysis_Plots import *

class Analysis_Module_tab(TabBase):

    def __init__(self, title='Modules'):
        super().__init__(title)

    def add_execute_btn(self, Network_UI, module):

        btn = Network_UI.Add_element(QPushButton('execute'), stretch=2)

        caption = Network_UI.Add_element(QLabel(str(module._get_base_name_())), stretch=4)
        spacing_blocks = 10

        argument_edits={}

        for attr_key in module.execution_arguments:
            attr = module.execution_arguments[attr_key]
            Network_UI.Add_element(QLabel(attr_key), stretch=2)
            edit=Network_UI.Add_element(QLineEdit(str(attr)), stretch=1)
            argument_edits[attr_key]=edit
            spacing_blocks -= 3

        if spacing_blocks > 0:
            Network_UI.Add_element(QLabel(''), stretch=spacing_blocks)

        def execute_clicked(event):
            Network_UI.pause = True
            module.add_progress_update_function(self.update_progress)
            arguments = {}
            for key in argument_edits:
                try:
                    arguments[key] = eval(argument_edits[key].text())
                except:
                    arguments[key] = argument_edits[key].text()

            module(**arguments)
            module.add_progress_update_function(None)


        btn.setEnabled(module.is_executable())
        btn.clicked.connect(execute_clicked)

    def add_result_remove_menu(self, Network_UI, module):

        if module.is_executable():
            cb = Network_UI.Add_element(Analytics_Results_Select_ComboBox(module), stretch=2)

            def remove_selected_clicked(event):
                key = cb.currentText()
                module.remove_result(key)

            btn = Network_UI.Add_element(QPushButton('remove result'), stretch=1)
            btn.clicked.connect(remove_selected_clicked)
        else:
            Network_UI.Add_element(QLabel(''), stretch=3)

    def update_progress(self, percentage):
        QApplication.instance().processEvents()
        self.progressbar.setValue(percentage)

    def initialize(self, Network_UI):
        self.a_module_tab = Network_UI.Next_Tab(self.title)

        print(Network_UI.network.NeuronGroups[0].analysis_modules)

        for obj in [Network_UI.network]+Network_UI.network.NeuronGroups:#Network_UI.network.all_objects()
            caption = Network_UI.Add_element(QLabel(obj.tags[0]))
            font = QFont('SansSerif', 16)
            font.setUnderline(True)
            caption.setFont(font)
            Network_UI.Next_H_Block()

            for analysis_module in obj.analysis_modules:

                self.add_execute_btn(Network_UI, analysis_module)

                self.add_result_remove_menu(Network_UI, analysis_module)

                Network_UI.Next_H_Block()

        self.progressbar = Network_UI.Add_element(QProgressBar())



from PymoNNto.Exploration.Network_UI.TabBase import *

class buffer_tab(TabBase):

    def __init__(self, title='Buffer'):
        super().__init__(title)

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):
        self.buffer_tab = Network_UI.Next_Tab(self.title)

        buffer_keys = {}
        for group_tag in Network_UI.group_tags:
            for ng in Network_UI.network[group_tag]:
                if hasattr(ng, 'buffers'):
                    for variable in ng.buffers:
                        for timescale in ng.buffers[variable]:
                            for offset in ng.buffers[variable][timescale]:
                                buffer_keys[str(variable)+' '+str(timescale)+' '+str(offset)]=1

        self.buffer_block_plot_images = []
        self.buffer_block_count=len(list(buffer_keys.keys()))
        for i in range(self.buffer_block_count):
            #Network_UI.Add_element(QLabel(''))
            image_item = Network_UI.Add_Image_Item(True, False, title='', tooltip_message='')
            self.buffer_block_plot_images.append(image_item)
            Network_UI.Next_H_Block()

    def update(self, Network_UI):
        if self.buffer_tab.isVisible():

            sub_group = Network_UI.get_selected_neuron_subgroup()
            if hasattr(sub_group, 'buffers'):
                ct = 0

                if sub_group is not None:
                    for variable in sub_group.buffers:
                        for timescale in sub_group.buffers[variable]:
                            for offset in sub_group.buffers[variable][timescale]:
                                b=sub_group.get_buffer(sub_group, variable, timescale, offset)
                                self.buffer_block_plot_images[ct][1].setLabels(title=str(variable)+' '+str(timescale)+' '+str(offset))
                                self.buffer_block_plot_images[ct][0].setImage(np.rot90(b, 3), levels=(0, 1))
                                ct+=1

                for i in range(self.buffer_block_count-ct):
                    self.buffer_block_plot_images[ct][0].clear()



from PymoNNto.Exploration.StorageManager.StorageManager import *
from PyQt5.QtWidgets import *
import sys
from PyQt5.QtGui import *
import PymoNNto
from PyQt5 import QtCore


class SM_D_S_Dialog(QDialog):

    def __init__(self, parent=None):
        super(SM_D_S_Dialog, self).__init__(parent)
        layout = QVBoxLayout()
        self.setLayout(layout)

        path = PymoNNto.__file__.replace('__init__.py', '')
        if 'win' in sys.platform and sys.platform != 'darwin':
            import ctypes
            myappid = 'mv.Pymonnto.ui.1'  # arbitrary string mycompany.myproduct.subproduct.version
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

        app_icon = QIcon()
        app_icon.addFile(path + 'icon3232.png', QtCore.QSize(32, 32))
        self.setWindowIcon(app_icon)

        self.listWidget=QListWidget()

        # Trying to set up the listWidgetView
        #self.listWidget.setEditTriggers(QtGui.QAbstractItemView.DoubleClicked | QtGui.QAbstractItemView.EditKeyPressed)
        self.listWidget.setSelectionMode(QAbstractItemView.MultiSelection)
        #self.listWidget.setViewMode(QtGui.QListView.ListMode)

        # Load the layers in the listWidget


        main_dir = get_data_folder() + '/StorageManager/'

        #os.path.join(main_dir, o)

        layers = [o for o in os.listdir(main_dir) if os.path.isdir(os.path.join(main_dir, o))]

        # Populate the listWidget with all the polygon layer present in the TOC
        self.listWidget.addItems(layers)

        layout.addWidget(self.listWidget)
        # put the selected layers in a list
        #selectedLayers = self.listWidget.selectedItems()



        # Add button signal to greetings slot
        self.button = QPushButton("Ok")
        self.button.clicked.connect(self.greetings)
        layout.addWidget(self.button)

        self.list=[]

        self.width = 800
        self.height = 600
        self.setWindowTitle('Select Folders')
        self.setGeometry(10, 40, self.width, self.height)

    def cancel(self):
        self.close()

    def greetings(self):
        self.list = [selected.text() for selected in self.listWidget.selectedItems()]
        self.accept()

    #def exec_(self):
    #    if QDialog.exec_(self) == QDialog.Accepted:
    #        return self.list
    #    else:
    #        return []

def Storage_Manager_Dir_Select_Dialog():
    app = QApplication(sys.argv)
    form = SM_D_S_Dialog()
    form.show()
    app.exec_()
    return form.list

if __name__ == '__main__':
    print(Storage_Manager_Dir_Select_Dialog())


from PymoNNto.Exploration.Network_UI.Advanced_Tabs.stdp_buffer_tab import *
from PymoNNto.Exploration.Network_UI.Advanced_Tabs.criticality_tab import *
from PymoNNto.Exploration.Network_UI.Advanced_Tabs.buffer_tab import *
from PymoNNto.Exploration.Network_UI.Advanced_Tabs.scatter_tab import *
from PymoNNto.Exploration.Network_UI.Advanced_Tabs.afferent_syn_attr_plot_tab import *

def get_advanced_UI_modules(neuron_parameters=['output'], synapse_parameters=['W']):
    return [
    stdp_buffer_tab(),
    criticality_tab(),
    buffer_tab(),
    scatter_tab(x_var='excitation', y_var='inhibition'),
    afferent_syn_attr_plot_tab(syn_vars=['slow_add', 'fast_add']),
]


from PymoNNto.Exploration.Evolution.Devices.Evolution_Device_Single_Thread import *
from PymoNNto.Exploration.Evolution.Devices.Evolution_Device_SSH import *
from PymoNNto.Exploration.Evolution.Breed_And_Selection_Module import *
from PymoNNto.Exploration.Evolution.UI_Single_Evolution_Monitor import *
import os.path
import time


class Evolution:

    def __init__(self, name, slave_file, individual_count=10, mutation=0.4, death_rate=0.5, devices={'single_thread':1}, constraints=[], start_genomes=[], inactive_genome_info={}, breed_and_select_module=Default_Breed_And_Select, additional_evo_params={}):

        self.Breed_And_Select = breed_and_select_module(self, death_rate=death_rate, mutation=mutation, individual_count=individual_count, constraints=constraints, additional_evo_params=additional_evo_params)

        self.name = name

        self.slave_file = slave_file

        self.devices = []
        self.start_genomes = start_genomes
        self.inactive_genome_info = inactive_genome_info

        self.id_counter = 0

        if not os.path.isfile(slave_file):
            print('warning slave file not found')

        if start_genomes==[]:
            print('Error no genomes found')

        for key in ['gen', 'score', 'id']:
            for genome in self.start_genomes:
                if key in genome:
                    genome.pop(key)
                    print(key, 'removed form genomes')


        self.scored_individuals = []
        self.running_individuals = []

        self.gene_keys = []
        if len(self.start_genomes) > 0:
            self.gene_keys = list(self.start_genomes[0].keys())
            for genome in self.start_genomes:
                if len(self.gene_keys) != len(genome):
                    print('Error: start genomes do not have the same size')
                for gene in genome:
                    if gene not in self.gene_keys:
                        print('Error: start genomes have different gene keys')

                if not self.Breed_And_Select.is_valid_genome(genome):
                    print('Error: start genome does not fulfill constraints', genome)

            self.non_scored_individuals = self.Breed_And_Select.breed(start_genomes)
            print(self.non_scored_individuals)
        else:
            print('Error no start genomes found')

        for device_string, number_of_threads in devices.items():
            self.add_devices(device_string, number_of_threads)

        if is_invalid_evo_name(name):
            print('Error: For savety reasons some names and characters are forbidden to avoid the accidental removal of files or folders')
            self.devices = {}

        for device in self.devices:
            device.initialize()

        self.ui = None

        return

    def iteration(self):
        for device in self.devices:
            device.main_loop_update()
        #time.sleep(1.0)

    def part_of_genome(self, small_genome, big_genome):
        result = True
        for key in small_genome:
            if key!='score' and key not in big_genome or small_genome[key] != big_genome[key]:
                result = False
        return result

    def find(self, genome_list, genome):
        found = None
        for g in genome_list:
            if self.part_of_genome(g, genome):
                found = g
        return found

    def new_score_event(self, genome):#called by devices
        found = self.find(self.running_individuals, genome)

        if found is not None:
            self.running_individuals.remove(found)
            found['score'] = genome['score']
            self.scored_individuals.append(found)
            self.Breed_And_Select.update_population()
            print('+', genome)#g

            if self.ui is not None:
                self.ui.update()

        else:
            self.error_event(genome, 'processed gene not found in running_individuals | running:' + str(self.running_individuals)+' scored:'+str(self.scored_individuals))

    def error_event(self, genome, message):#called by devices
        print('failed', message, genome)

        found = None
        for g in self.running_individuals:
            if self.part_of_genome(g, genome):
                found = g

        if found is not None:
            self.running_individuals.remove(found)
            self.non_scored_individuals.append(found)
            print('move genome from running_individuals back to non_scored_individuals...')
        else:
            print('not able to find failed genome in running_individuals. Maybe the genome was changed during processing')


    def add_device(self, device_string):
        if device_string == 'single_thread':
            self.devices.append(Evolution_Device_Single_Thread(device_string, self))
        if device_string == 'multi_thread':
            self.devices.append(Evolution_Device_Multi_Thread(device_string, self))
        if 'ssh' in device_string:
            self.devices.append(Evolution_Device_SSH(device_string, self))

    def add_devices(self, device_string, number_of_threads):
        # move folder to remote device if needed
        for _ in range(number_of_threads):
            self.add_device(device_string)

        if len(self.devices) > 0:
            self.devices[-1].initialize_device_group()

    def _run_evo(self, ui=True):
        for device in self.devices:
            device.start()

        if ui:
            self.ui = UI_Single_Evolution_Monitor(self)
            self.ui.show()
        else:
            self.active = True
            while self.active:
                for device in self.devices:
                    device.main_loop_update()
                time.sleep(0.01)

        for device in self.devices:
            device.stop()

    def continue_evolution(self, ui=True):

        genomes = []

        smg = StorageManagerGroup(self.name)

        smg.sort_by('gen')
        gens = smg.get_param_list('gen', remove_None=True)
        last_full_gen = np.max(gens)-1
        for sm in smg['gen=='+str(last_full_gen)]:
            genome = {}
            try:
                valid=True
                for key in self.gene_keys:
                    genome[key] = sm.load_param(key)
                    if genome[key] is None:
                        valid = False
                genome['score'] = sm.load_param('score')
                if genome['score'] is not None and valid:
                    genomes.append(genome)
            except:
                print('failed to load gene from', sm.folder_name)

        ids = smg.get_param_list('id', remove_None=True)
        max_id = np.max(ids)
        self.id_counter = max_id+1

        if len(genomes)!=0:
            self.scored_individuals = genomes
            self.running_individuals = []
            self.non_scored_individuals = []
            self.Breed_And_Select.generation = last_full_gen

        print('Continuing', self.name, 'at generation', last_full_gen, 'with the following genomes:')
        print(self.scored_individuals)

        self.Breed_And_Select.update_population()

        self._run_evo(ui)

    def start(self, ui=True):
        folder = get_data_folder()+'/StorageManager/'+self.name
        if not os.path.isdir(folder):

            self._run_evo(ui)
            return True
        else:
            print('Warning:', folder, 'already exists. Remove folder or try continue_evolution() instead of start()')
            return False

    def stop(self):
        for device in self.devices:
            device.stop()

    def add_name_gen_id_inactive_to_genome(self, genome):
        result = genome.copy()
        result['evo_name'] = self.name
        result['gen'] = self.Breed_And_Select.generation
        result['id'] = self.id_counter
        self.id_counter += 1
        result.update(self.inactive_genome_info)
        return result

    def get_next_genome(self):
        result = None
        if len(self.non_scored_individuals) > 0:
            result = self.non_scored_individuals[0]
            self.non_scored_individuals.pop(0)
            self.running_individuals.append(result)

        #if result is None and len(self.running_individuals) > 0:
        #    result = self.running_individuals[0]

        if result is not None:
            result = self.add_name_gen_id_inactive_to_genome(result)
            #result = result.copy()
            #result['evo_name'] = self.name
            #result['gen'] = self.Breed_And_Select.generation
            #result['id'] = self.id_counter
            #self.id_counter += 1
            #result.update(self.inactive_genome_info)

        return result


from PymoNNto.NetworkCore.Analysis_Module import *

class Neuron_Classification_Colorizer(AnalysisModule):

    def initialize(self, neurons):
        self.add_tag('colorizer')
        self.unique_colors = {}
        self.compiled_formats = {}

    def get_color_list(self, neuron_class_list, format='[r,g,b,255.0]'):
        if neuron_class_list is None:
            neuron_class_list = np.zeros(self.parent.size)
        color_dict = self.get_class_color_dict(neuron_class_list, format)
        return [color_dict[neuron_class_list[c]] for c in range(self.parent.size)]

    def get_class_color_dict(self, neuron_class_list, format='[r,g,b,255.0]'):
        unique = np.unique(neuron_class_list)
        colors = self.get_unique_colors(len(unique), format=format)
        return dict(zip(unique, colors))

    def get_unique_colors(self, number, format='[r,g,b,255.0]'):#[r,g,b,a]
        if not format in self.unique_colors:
            self.unique_colors[format] = []
            self.compiled_formats[format] = compile(format, '<string>', 'eval')

        comp = self.compiled_formats[format]
        rnd_r = 'r' in format
        rnd_g = 'g' in format
        rnd_b = 'b' in format
        rnd_a = 'a' in format

        while len(self.unique_colors[format]) < number:
            if rnd_r: r = np.random.rand()*255.0
            if rnd_g: g = np.random.rand()*100.0 #lower because green is the UI base color
            if rnd_b: b = np.random.rand()*255.0
            if rnd_a: a = np.random.rand()*255.0

            if hasattr(self.parent, 'color'):
                if len(self.unique_colors[format]) == 0:
                    r = self.parent.color[0]
                    g = self.parent.color[1]
                    b = self.parent.color[2]
                if len(self.unique_colors[format]) == 1:
                    r = self.parent.color[0]*0.6
                    g = self.parent.color[1]*0.6
                    b = self.parent.color[2]*0.6
                if len(self.unique_colors[format]) == 2:
                    r = self.parent.color[0]*0.3
                    g = self.parent.color[1]*0.3
                    b = self.parent.color[2]*0.3

            self.unique_colors[format].append(eval(comp))
        return self.unique_colors[format]



from PymoNNto.Exploration.Evolution.Devices.Evolution_Device import *

class Evolution_Device_Single_Thread(Evolution_Device):

    def get_score(self, genome):
        sm = StorageManager(main_folder_name=genome['evo_name'], folder_name=get_gene_file(genome), add_new_when_exists=False)#, print_msg=False
        return sm.load_param('score', default=None)

    def main_loop_update(self):
        current_genome = self.parent.get_next_genome()

        if current_genome is not None:

            if os.path.isfile(self.parent.slave_file):
                execute_local_file(self.parent.slave_file, current_genome)
            else:
                print('file not found')

            self.score_processing(current_genome)

        else:
            print('no genes found to process')



from PymoNNto.NetworkCore.Behaviour import *
import copy

def get_Recorder(variable):
    return Recorder(variables=[variable])


class Recorder(Behaviour):
    visualization_module_outputs = []

    def __init__(self, variables, gapwidth=0, tag=None, max_length=None, save_as_numpy=False):
        super().__init__(tag=tag, variables=variables, gapwidth=gapwidth, max_length=max_length, save_as_numpy=save_as_numpy)
        #if tag is not None:
        #    self.add_tag(tag)
        self.add_tag('recorder')

        self.gapwidth = self.get_init_attr('gapwidth', 0)
        self.counter = 0
        self.new_data_available=False

        if type(variables) is str:
            variables=[variables]

        self.variables = {}
        self.compiled = {}
        self.save_as_numpy = self.get_init_attr('save_as_numpy', False)

        self.add_variables(self.get_init_attr('variables', []))
        self.reset()
        self.max_length = self.get_init_attr('max_length', None)

    def set_variables(self, neurons):
        self.reset()

    def add_varable(self, v):
        if self.save_as_numpy:
            self.variables[v] = np.array([])
        else:
            self.variables[v] = []
        self.compiled[v] = None

    def add_variables(self, vars):
        for v in vars:
            self.add_varable(v)

    def find_objects(self, key):
        result = []
        if key in self.variables:
            result.append(self.variables[key])
        return result

    def reset(self):
        for v in self.variables:
            self.add_varable(v)



    def is_new_data_available(self):
        if self.new_data_available:
            self.new_data_available = False
            return True
        else:
            return False

    def get_data_v(self, variable, parent_obj):
        n = parent_obj  # used for eval string "n.x"
        s = parent_obj
        neurons = parent_obj
        synapse = parent_obj

        return copy.copy(eval(self.compiled[variable]))

    def save_data_v(self, data, variable):
        if self.save_as_numpy:
            if len(self.variables[variable]) == 0:
                self.variables[variable] = np.array([data])
            else:
                self.variables[variable] = np.concatenate([self.variables[variable], [data]], axis=0)
        else:
            self.variables[variable].append(data)

    def new_iteration(self, parent_obj):
        if parent_obj.recording:

            self.counter += 1
            if self.counter >= self.gapwidth:
                self.new_data_available = True
                self.counter = 0
                for v in self.variables:
                    if self.compiled[v] is None:
                        self.compiled[v] = compile(v, '<string>', 'eval')
                    data = self.get_data_v(v, parent_obj)
                    if data is not None:
                        self.save_data_v(data, v)

        if self.max_length is not None:
            self.cut_length(self.max_length)

    def cut_length(self, max_length):
        if max_length is not None:
            for v in self.variables:
                while len(self.variables[v]) > max_length:
                    if self.save_as_numpy:
                        print('not implemented')
                    else:
                        self.variables[v].pop(0)

    def swaped(self, name):
        return self.swap(self.variables[name])#getattr(self, name)

    def swap(self, x):
        return np.swapaxes(np.array(x), 1, 0)

    def clear_recorder(self):
        print('clear')
        for v in self.variables:
            self.add_varable(v)

class EventRecorder(Recorder):

    def __init__(self, variables, tag=None):
        super().__init__(variables, gapwidth=0, tag=tag, max_length=None, save_as_numpy=True)

    def find_objects(self, key):

        result = []
        if key in self.variables:
            result.append(self.variables[key])

        if type(key) is str and key[-2:] == '.t' and key[:-2] in self.variables:
            result.append(self.variables[key[:-2]][:, 0])

        if type(key) is str and key[-2:] == '.i' and key[:-2] in self.variables:
            result.append(self.variables[key[:-2]][:, 1])

        return result

    def get_data_v(self, variable, parent_obj):
        n = parent_obj  # used for eval string "n.x"
        s = parent_obj
        neurons = parent_obj
        synapse = parent_obj

        data = eval(self.compiled[variable])
        indices = np.where(data != 0)[0]

        if len(indices) > 0:
            result = []
            for i in indices:
                result.append([parent_obj.iteration, i])
            return result
        else:
            return None

    def save_data_v(self, data, variable):
        if len(self.variables[variable]) == 0:
            self.variables[variable] = np.array(data)
        else:
            self.variables[variable] = np.concatenate([self.variables[variable], data], axis=0)

'''
class SynapseGroupRecorder(Recorder):

    def __init__(self, variables, transmitter='GLU', gapwidth=0, tag=None, max_length=None):
        self.transmitter = transmitter
        super().__init__([], gapwidth, tag=tag, max_length=max_length)
        self.add_variables(variables)

    def find_objects(self, key):
        result = []
        k = self.get_synapse_command(key)
        if k in self.variables:
            result.append(self.variables[k])
        return result

    def add_variables(self, vars):
        super().add_variables([self.get_synapse_command(v) for v in vars])

    def get_synapse_command(self, var):
        return 'np.concatenate([' + var + ' for s in neurons.afferent_synapses.get("' + self.transmitter + '", np.array([]))], axis=1).flatten()'

    #def __getitem__(self, key):
    #    return np.array(self.variables[self.get_synapse_command(key)])
'''









'''

class TRENRecorder(Neuron_Behaviour):

    def __init__(self, variables, behaviour_index=None, gapwidth=0):
        super().__init__()
        self.gapwidth=gapwidth
        self.counter = 0
        self.variables = []
        self.indexes = []
        for var in variables:
            var, indx = self.split_attribute_and_index(var)
            self.variables.append(var)
            self.indexes.append(indx)
        self.behaviour_index = behaviour_index
        self.reset()

    def reset(self):
        for var in self.variables:
            setattr(self, var, [])

    def set_variables(self, neurons):
        self.reset()

    def new_iteration(self, neurons):
        self.counter+=1
        if self.counter>self.gapwidth:
            self.counter=0
            src = neurons
            if self.behaviour_index is not None:
                src = neurons.behaviour[self.behaviour_index]

            for var, indx in zip(self.variables, self.indexes):
                val = getattr(src, var).copy()

                if indx is None:
                    getattr(self, var).append(val)
                else:
                    getattr(self, var).append(val[indx])

    def swaped(self, name):
        return self.swap(getattr(self, name))

    def swap(self, x):
        return np.swapaxes(np.array(x), 1, 0)

    def split_attribute_and_index(self, name):
        if '[' in name:
            split = name.split('[')
            return split[0], int(split[1][:-1])
        else:
            return [name, None]


class TRENSynapseRecorder(TRENRecorder):

    def __init__(self, transmitter, variables, gapwidth=0):
        super().__init__(variables, gapwidth)
        self.transmitter = transmitter

    def set_variables(self, neurons):
        super().set_variables(neurons)
        syns = neurons.afferent_synapses.get(self.transmitter, [])
        start = 0
        self.block_sizes = []
        for s in syns:
            size = s.get_dest_size()*s.get_src_size()
            end = start+size
            self.block_sizes.append([start, end])
            start=end

    def get_block_num(self):
        return len(self.block_sizes)

    def get_block(self, index, mat):
        return np.array(mat)[:, self.block_sizes[index][0]:self.block_sizes[index][1]]

    def new_iteration(self, neurons):
        self.counter+=1
        if self.counter>self.gapwidth:
            self.counter=0
            for var in self.variables:
                syns = neurons.afferent_synapses.get(self.transmitter, [])

                vals=np.array([])
                for s in syns:
                    vals = np.concatenate((vals, getattr(s, var).flatten()))

                getattr(self, var).append(vals)
                
'''



from PymoNNto.Exploration.Network_UI.TabBase import *

#from Testing.Common.Classifier_Helper import *

class sidebar_drumbeat_module(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        if hasattr(neuron_group, 'pattern_index'):
            Network_UI.add_recording_variable(neuron_group, 'n.pattern_index', timesteps=100)

    def initialize(self, Network_UI):
        if len(Network_UI.network['drum_act'])>0:

            self.readout = None

            def learning_on_off(event):
                Network_UI.network.set_mechanisms(['STDP'], self.stdp_cb.isChecked())

            network = Network_UI.network
            source = network['drum_act', 0]

            p = Network_UI.Add_plot(sidebar=True, title='Input')#,axisItems={'left': stringaxis})

            self.scatter = pg.ScatterPlotItem()
            p.showGrid(y=True)
            p.addItem(self.scatter)

            p.setYRange(-1,source.A+2, padding=0)

            self.n_notes_shown=20
            p.setXRange(-1,self.n_notes_shown+1)


    def update(self, Network_UI):
        if len(Network_UI.network['drum_act'])>0:

            network = Network_UI.network
            source = network['drum_act', 0]
            curr_output = np.array(network['n.pattern_index', 0][-self.n_notes_shown:])

            y=[]
            x=[]
            ts=0
            for i in range(curr_output.shape[0]):
                indices = np.nonzero(curr_output[i])[0]
                if len(indices)==0:
                    x.append(ts)
                    y.append(np.nan)
                else:
                    for j in indices:
                        x.append(ts)
                        y.append(j)
                ts+=1
            self.scatter.setData(x,y)

            if len(x) > 0:#self.n_notes_shown:
                self.scatter.setData(x,y)
                self.scatter.setSize(2)

        



from PymoNNto.Exploration.Network_UI.TabBase import *
from functools import partial

class sidebar_fast_forward_module(TabBase):

    def __init__(self, step_list=[100, 1000, 5000, 15000, 30000, 50000, 70000, 100000], title='PCA'):
        super().__init__(title)
        self.step_list = step_list

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def recorder_on_off(self, Network_UI, enable):
        for rec in Network_UI.network['UI_rec']:
            rec.behaviour_enabled = enable or self.rec_cb.isChecked()

    def update_progress(self, progress):
        #self.progressbar.setHidden(False)
        QApplication.instance().processEvents()
        self.progressbar.setValue(progress)

    def fast_forward(self, steps, Network_UI):
        self.update_progress(0)
        Network_UI.timer.stop()
        self.recorder_on_off(Network_UI, False)
        Network_UI.network.simulate_iterations(steps, 100, measure_block_time=True, batch_progress_update_func=self.update_progress)
        self.recorder_on_off(Network_UI, True)
        Network_UI.timer.start()

    def initialize(self, Network_UI):
        self.iteration_display_label=Network_UI.Add_element(QLabel(),sidebar=True)
        self.iteration_display_label.setMaximumHeight(10)

        def start_pause_click(event):
            Network_UI.pause = not Network_UI.pause

        self.sp_btn = QPushButton('start/pause', Network_UI.main_window)
        self.sp_btn.clicked.connect(start_pause_click)
        Network_UI.Add_Sidebar_Element(self.sp_btn)

        h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)
        h_layout.addWidget(QLabel('Render every X frames.'))
        self.render_every_x_frames_spin = QSpinBox()
        self.render_every_x_frames_spin.setMinimum(1)
        self.render_every_x_frames_spin.setMaximum(1000)

        def change_rexf(event):
            Network_UI.render_every_x_frames = self.render_every_x_frames_spin.value()

        self.render_every_x_frames_spin.valueChanged.connect(change_rexf)

        h_layout.addWidget(self.render_every_x_frames_spin)

        if Network_UI.storage_manager is not None:
            self.record_frames_cb=QCheckBox('save frames')
            def cb_state_changed(event):
                if not self.record_frames_cb.isChecked():
                    dlg = QDialog()
                    dlg.setWindowTitle("Do you want to render a video?")
                    layout = QVBoxLayout()

                    path_tl=QLineEdit(Network_UI.storage_manager.absolute_path)
                    path_tl.setReadOnly(True)
                    layout.addWidget(path_tl)

                    delete_frames_cb = QCheckBox('delete frame images')
                    layout.addWidget(delete_frames_cb)

                    def render():
                        Network_UI.storage_manager.render_video('ui_frame_', delete_images=delete_frames_cb.isChecked())
                        dlg.close()


                    render_btn = QPushButton('render video')
                    render_btn.clicked.connect(render)

                    layout.addWidget(render_btn)
                    dlg.setLayout(layout)
                    dlg.resize(300, 50)
                    dlg.exec()

            self.record_frames_cb.stateChanged.connect(cb_state_changed)
            h_layout.addWidget(self.record_frames_cb)

        line=Network_UI.Add_element(QFrame(),sidebar=True)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)


        h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

        Network_UI.step = False

        def one_step(event):
            Network_UI.step = True

        self.os_btn = QPushButton('X steps', Network_UI.main_window)
        self.os_btn.clicked.connect(one_step)
        # self.Add_Sidebar_Element(self.os_btn)
        h_layout.addWidget(self.os_btn)


        for i, steps in enumerate(self.step_list):

            if (i+1)%3==0:
                h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

            def ff_btn_clicked(steps):
                self.fast_forward(steps, Network_UI)

            txt = str(steps)

            if steps%1000==0:
                txt=str(int(steps/1000))+'k'

            ff_btn = QPushButton(txt, Network_UI.main_window)
            ff_btn.clicked.connect(partial(ff_btn_clicked, steps))

            h_layout.addWidget(ff_btn)

        '''
        def fast_forward(event):
            self.fast_forward(100, Network_UI)

        self.ff_btn = QPushButton('100', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        def fast_forward(event):
            self.fast_forward(1000, Network_UI)

        self.ff_btn = QPushButton('1k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        h_layout = Network_UI.Add_Sidebar_Element(return_h_layout=True)

        def fast_forward(event):
            self.fast_forward(5000, Network_UI)

        self.ff_btn = QPushButton('5k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        def fast_forward(event):
            self.fast_forward(15000, Network_UI)

        self.ff_btn = QPushButton('15k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)

        def fast_forward(event):
            self.fast_forward(50000, Network_UI)

        self.ff_btn = QPushButton('50k', Network_UI.main_window)
        self.ff_btn.clicked.connect(fast_forward)
        # self.Add_Sidebar_Element(self.ff_btn)
        h_layout.addWidget(self.ff_btn)
        '''

        self.rec_cb = QCheckBox()
        self.rec_cb.setText('fast forward record')
        self.rec_cb.setChecked(False)
        Network_UI.Add_Sidebar_Element(self.rec_cb)

        self.progressbar, = Network_UI.Add_Sidebar_Element(QProgressBar())
        #self.progressbar.setHidden(True)

        line=Network_UI.Add_element(QFrame(),sidebar=True)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)


    def record_frame(self, Network_UI):
        if Network_UI.storage_manager is not None:
            pix = QtGui.QPixmap(Network_UI.main_window.size())
            Network_UI.main_window.render(pix)
            file_path=Network_UI.storage_manager.get_next_frame_name('ui_frame_')
            pix.save(file_path)

    def update(self, Network_UI):

        self.iteration_display_label.setText('Iteration: '+str(Network_UI.network.iteration))

        if Network_UI.storage_manager is not None and self.record_frames_cb.isChecked():
            self.record_frame(Network_UI)



from PymoNNto.NetworkBehaviour.Recorder.Recorder import *
import copy

class NeuronGroup(NetworkObjectBase):

    def __init__(self, size, behaviour, net, tag=None):
        super().__init__(tag=tag)
        self.add_tag('ng')

        self.BaseNeuronGroup = self#used for subgroup reconstruction

        if isinstance(size, Behaviour):
            if type(behaviour) is dict:
                if 0 in behaviour:
                    print('warning: 0 index behaviour will be overwritten by size behaviour')
                behaviour[0] = size
            if type(behaviour) is list:
                behaviour.append(size)#first index???
            size = -1

        if net is not None:
            net.NeuronGroups.append(self)
            self.network = net


        self.size = size

        self.afferent_synapses = {} #set by Network
        self.efferent_synapses = {}

        self.mask = True#np.array([True for _ in range(size)]).astype(np.bool)#True#used for subgroup reconstruction

        self.learning = True
        self.recording = True

        self.behaviour = behaviour #Todo: Test without copy
        #if type(behaviour) == list:
        #    for i, b in enumerate(behaviour):
        #        self.behaviour[i] = copy.copy(b)
        #else:
        #    for k in behaviour:
        #        self.behaviour[k] = copy.copy(behaviour[k])

        for k in sorted(list(self.behaviour.keys())):
            if self.behaviour[k].set_variables_on_init:
                self.behaviour[k].set_variables(self)

        self.id = np.arange(self.size)

    #do only use before initialization!!!
    #after initialization use Netork.add_behaviours_to_neuron_group(...)
    def add_behaviour(self, index, b, init=False):
        self.behaviour[index] = copy.copy(b)
        if init:
            self.behaviour[index].set_variables(self)
            self.behaviour[index].check_unused_attrs()
        return self.behaviour[index]

    def find_objects(self, key):
        result = []

        if key in self.behaviour:
            result.append(self.behaviour[key])

        for bk in self.behaviour:
            behaviour = self.behaviour[bk]
            result += behaviour[key]

        #for syn_key in self.afferent_synapses:
        #    for syn in self.afferent_synapses[syn_key]:
        #        if syn not in result:
        #            result += syn[key]

        for am in self.analysis_modules:
            result += am[key]

        return result

    def get_modification_value(self, key):
        result = getattr(self, key)
        if type(result) == np.ndarray:
            return np.sum(result)
        return result

    def set_modification_value(self, key, value):
        old_val = getattr(self, key)
        if type(old_val) == np.ndarray:
            mat_sum = np.sum(old_val)
            if mat_sum == 0.0:
                new = np.ones(old_val.shape)*value
            else:
                new = old_val/mat_sum*value
            print('mat set to:', new)
            return setattr(self, key, new)
        else:
            return setattr(self, key, value)

    def require_synapses(self, name, afferent=True, efferent=True, warning=True):
        if afferent and not name in self.afferent_synapses:
            if warning:
                print('warning: no afferent {} synapses found'.format(name))
            self.afferent_synapses[name] = []

        if efferent and not name in self.efferent_synapses:
            if warning:
                print('warning: no efferent {} synapses found'.format(name))
            self.efferent_synapses[name] = []

    #def get_neuron_vec(self):
    #    return self.get_nparray((self.size))

    def get_neuron_vec(self, mode='zeros()', scale=None, density=None, plot=False, kwargs={}, args=[]):# mode in ['zeros', 'zeros()', 'ones', 'ones()', 'uniform(...)', 'lognormal(...)', 'normal(...)', ...]
        return self._get_mat(mode=mode, dim=(self.size), scale=scale, density=density, plot=plot, kwargs=kwargs, args=args)

    def get_neuron_vec_buffer(self, buffer_size):
        return self.get_buffer_mat((self.size), buffer_size)

    #def get_random_neuron_vec(self, density=1.0):#r
    #    print("warning: use get_neuron_vec('uniform',...) instead of get_random_neuron_vec")
    #    return self.get_random_nparray((self.size), density)

    #def get_shared_variables(self, name, avoid_None=True):
    #    result=[]
    #    for k in self.behaviour:
    #        val = self.behaviour[k].get_shared_variable(name)
    #        if val!=None or avoid_None==False:
    #            result.append(val)
    #    return np.array(result)

    def get_combined_synapse_shape(self, Synapse_ID):
        source_num = 0
        for syn in self.afferent_synapses[Synapse_ID]:
            d, s = syn.get_synapse_mat_dim()
            source_num += s
        return self.size, source_num

    def __str__(self):
        result = 'NeuronGroup'+str(self.tags)+'('+str(self.size)+'){'
        for k in sorted(list(self.behaviour.keys())):
            result += str(k) + ':' + str(self.behaviour[k])
        return result+'}'

    def subGroup(self, mask=None):
        return NeuronSubGroup(self, mask)

    def group_without_subGroup(self):
        return self

    def get_masked_dict(self, dict_name, key):
        return getattr(self, dict_name)[key]

    def connected_NG_param_list(self, param_name, syn_tag='All', afferent_NGs=False, efferent_NGs=False, same_NG=False, search_behaviours=False):
        result = []

        def search_NG(NG):
            if hasattr(NG, param_name):
                attr = getattr(NG, param_name)
                if callable(attr):
                    result.append(attr(NG))
                else:
                    result.append(attr)
            if search_behaviours:
                for key, behaviour in NG.behaviour.items():
                    if hasattr(behaviour, param_name):
                        attr = getattr(behaviour, param_name)
                        if callable(attr):
                            result.append(attr(NG))
                        else:
                            result.append(attr)

        if same_NG:
            search_NG(self)

        if efferent_NGs:
            for syn in self.efferent_synapses[syn_tag]:
                search_NG(syn.dst)

        if afferent_NGs:
            for syn in self.afferent_synapses[syn_tag]:
                search_NG(syn.src)

        return result

    def partition_size(self, block_size=7):

        w = block_size#int((self.src.width/block_size+self.dst.width/block_size)/2)
        h = block_size#int((self.src.height/block_size+self.dst.height/block_size)/2)
        d = block_size#int((self.src.depth / block_size + self.dst.depth / block_size) / 2)
        split_size = [np.maximum(w, 1), np.maximum(h, 1), np.maximum(d, 1)]
        if split_size[0]<2 and split_size[1]<2 and split_size[2]<2:
            return [self]
        else:
            return self.split_grid_into_sub_group_blocks(split_size)

    def partition_masks(self, steps=[1, 1, 1]):

        dst_min = [np.min(p) for p in [self.x, self.y, self.z]]
        dst_max = [np.max(p) for p in [self.x, self.y, self.z]]

        def get_start_end(step, dim):
            start=dst_min[dim]+(dst_max[dim]-dst_min[dim])/steps[dim]*step
            end=dst_min[dim]+(dst_max[dim]-dst_min[dim])/steps[dim]*(step+1)
            return start, end

        results = []

        masks = []
        for w_step in range(steps[0]):          #x_steps
            dst_x_start, dst_x_end = get_start_end(w_step, 0)
            for h_step in range(steps[1]):      #y_steps
                dst_y_start, dst_y_end = get_start_end(h_step, 1)
                for d_step in range(steps[2]):  #z_steps
                    dst_z_start, dst_z_end = get_start_end(d_step, 2)

                    sub_group_mask = (self.x >= dst_x_start) * (self.x <= dst_x_end) * (self.y >= dst_y_start) * (self.y <= dst_y_end) * (self.z >= dst_z_start) * (self.z <= dst_z_end)

                    #remove redundancies
                    for old_dst_mask in masks:
                        sub_group_mask[old_dst_mask] *= False
                    masks.append(sub_group_mask)

                    results.append(sub_group_mask)

        return results

    def split_grid_into_sub_group_blocks(self, steps=[1, 1, 1]):
        return [self.subGroup(mask) for mask in self.partition_masks(steps)]


    def get_subgroup_receptive_field_mask(self, subgroup, xyz_rf=[1,1,1]):
        rf_x, rf_y, rf_z = xyz_rf

        src_x_start = np.min(subgroup.x) - rf_x
        src_x_end = np.max(subgroup.x) + rf_x

        src_y_start = np.min(subgroup.y) - rf_y
        src_y_end = np.max(subgroup.y) + rf_y

        src_z_start = np.min(subgroup.z) - rf_z
        src_z_end = np.max(subgroup.z) + rf_z

        mask = (self.x >= src_x_start) * (self.x <= src_x_end) * (self.y >= src_y_start) * (self.y <= src_y_end) * (self.z >= src_z_start) * (self.z <= src_z_end)

        return mask

    def mask_var(self, var):
        return var

    def visualize_module(self, show_dimension=True, show_recorder=True, only_feed_forward_connections=False):
        from PymoNNto.Exploration.Visualization import Module_visualizer as drawer
        md = drawer.module_drawer()

        for k in sorted(list(self.behaviour.keys())):
            b = self.behaviour[k]
            b.set_gene_variables()
            if (b.__class__.__name__ != 'Recorder' or show_recorder) and (b.__class__.__name__ != 'NeuronDimension' or show_dimension):
                md.add_module(b)

        md.add_flow_chart(only_feed_forward_connections)

        md.show()

    '''
    def get_neuron_behaviour_parameter_list(self, neurons, param):
        result = []
        for k, b in neurons.behaviour.items():
            if hasattr(b, param):
                result.append(getattr(b, param))
        return result

    def get_connected_neuron_parameter_list(self, neurons, param, syn_tag='GLU', afferent=False, efferent=False, same_ng=False, search_behaviours=False):
        result = []
        if efferent:
            for syn in neurons.efferent_synapses[syn_tag]:
                if hasattr(syn.dst, param):
                    result.append(getattr(syn.dst, param))

        if afferent:
            for syn in neurons.afferent_synapses[syn_tag]:
                if hasattr(syn.src, param):
                    result.append(getattr(syn.src, param))

        return result
            #max_buf = max(max_buf, syn.dst.input_buffer_requirement)
        #neurons.output_buffer = neurons.get_neuron_vec_buffer(max_buf)
    '''


class NeuronSubGroup:

    def __init__(self, BaseNeuronGroup, mask):
        self.cache = {}
        self.key_id_cache = {}
        self.BaseNeuronGroup = BaseNeuronGroup
        self.mask = mask
        self.id_mask = np.where(mask)[0]

    def mask_var(self, var):
        if var.shape[0] != self.mask.shape[0]:
            return var[:, self.mask]
        else:
            return var[self.mask]

    #def get_masked_dict(self, dict_name, key):
    #    attr=getattr(self.BaseNeuronGroup, dict_name)[key]
    #    self.mask_var(attr)
    #    #if attr.shape[0] != self.mask.shape[0]:
    #    #    return attr[:, self.mask]
    #    #else:
    #    #    return attr[self.mask]

    def __getattr__(self, attr_name):
        if attr_name in ['BaseNeuronGroup', 'mask', 'cache', 'key_id_cache', 'id_mask']:
            return super().__getattr__(attr_name)#setattr

        if attr_name == 'size':
            return np.sum(self.mask)

        #print('request', attr_name,hasattr(self.BaseNeuronGroup, 'tf'))

        attr = getattr(self.BaseNeuronGroup, attr_name)
        if type(attr) == np.ndarray:

            #attr_id = id(attr)
            #if not attr_id in self.cache:
            #    if attr_name in self.key_id_cache:#remove unused cache data
            #        self.cache.pop(self.key_id_cache[attr_name])
            #    if attr.shape[0] != self.mask.shape[0]:
            #        self.cache[attr_id] = attr[:, self.mask]
            #    else:
            #        self.cache[attr_id] = attr[self.mask]
            #    self.key_id_cache[attr_name] = attr_id
            #return self.cache[attr_id]

            #self.mask_var(attr)

            if attr.shape[0] != self.mask.shape[0]:
                return attr[:, self.mask]
            else:
                return attr[self.mask]


        elif hasattr(self.BaseNeuronGroup, 'tf') and (type(attr) == self.BaseNeuronGroup.tf.python.ops.resource_variable_ops.ResourceVariable or type(attr) == self.BaseNeuronGroup.tf.python.framework.ops.EagerTensor):

            if attr.shape[0] != self.mask.shape[0]:
                return self.BaseNeuronGroup.tf.boolean_mask(attr, self.mask, axis=1)#attr[:, self.mask]
            else:
                return self.BaseNeuronGroup.tf.boolean_mask(attr, self.mask, axis=0)#attr[self.mask]
        else:
            return attr

    def __setattr__(self, attr_name, value):
        if attr_name in ['BaseNeuronGroup', 'mask', 'cache', 'key_id_cache', 'id_mask']:
            super().__setattr__(attr_name, value)
            return

        #if len(self.BaseNeuronGroup['TF']) > 0 and not hasattr(self, 'tf'):
        #    import tensorflow as tf
        #    self.tf=tf

        #print(self.BaseNeuronGroup.__dict__)

        attr = getattr(self.BaseNeuronGroup, attr_name)
        if type(attr) == np.ndarray:

            #attr_id = id(attr)
            #if not attr_id in self.cache:
            #    if attr_name in self.key_id_cache:  # remove unused cache data
            #        self.cache.pop(self.key_id_cache[attr_name])
            #    if attr.shape[0] != self.mask.shape[0]:
            #        self.cache[attr_id] = attr[:, self.mask]
            #    else:
            #        self.cache[attr_id] = attr[self.mask]
            #    self.key_id_cache[attr_name] = attr_id
            #self.cache[attr_id] = value

            #self.mask_var(attr)[:] = value

            if attr.shape[0] != self.mask.shape[0]:
                attr[:, self.mask] = value
            else:
                attr[self.mask] = value

        elif hasattr(self.BaseNeuronGroup, 'tf') and (type(attr) == self.BaseNeuronGroup.tf.python.ops.resource_variable_ops.ResourceVariable or type(attr) == self.BaseNeuronGroup.tf.python.framework.ops.EagerTensor):
            if attr.shape[0] != self.mask.shape[0]:
                self.BaseNeuronGroup.tf.compat.v1.scatter_update(attr, [np.newaxis, self.id_mask], value)
                #setattr(self.BaseNeuronGroup, attr_name, self.BaseNeuronGroup.tf.compat.v1.scatter_update(attr, [np.newaxis, self.id_mask], value))
            else:
                #setattr(self.BaseNeuronGroup, attr_name, self.BaseNeuronGroup.tf.compat.v1.scatter_update(attr, [self.id_mask], value))
                self.BaseNeuronGroup.tf.compat.v1.scatter_update(attr, [self.id_mask], value)


        else:
            setattr(self.BaseNeuronGroup, attr, value)

    def group_without_subGroup(self):
        return self.BaseNeuronGroup



from PymoNNto.Exploration.Network_UI.TabBase import *
import random

class single_group_plot_tab(TabBase):

    def __init__(self, variables=['output', 'excitation', 'inhibition'], colors=[(0, 0, 0), (0, 0, 255), (255, 0, 0), (155, 0, 155), (0, 155, 0), (155, 155, 0), (0, 155, 155)], net_lines=[], neuron_lines=[], title='Single Group', timesteps=500):
        super().__init__(title)
        self.variables = {}
        self.colors = colors
        self.net_lines = net_lines
        self.neuron_lines = neuron_lines
        for i, v in enumerate(variables):
            self.variables[v] = self.get_color(i)
        self.timesteps=timesteps

    def get_color(self, i):
        if i < len(self.colors):
            return self.colors[i]
        else:
            return (random.randint(0,255),random.randint(0,255),random.randint(0,255))


    def add_recorder_variables(self, neuron_group, Network_UI):

        for var in self.variables:
            if hasattr(neuron_group, var):
                Network_UI.add_recording_variable(neuron_group, 'n.'+var, timesteps=self.timesteps)
                Network_UI.add_recording_variable(neuron_group, 'np.mean(n.'+var+')', timesteps=self.timesteps)

    def initialize(self, Network_UI):
        self.inh_exc_tab = Network_UI.Next_Tab(self.title, stretch=0)

        self.group_title_label = Network_UI.Add_element(QLabel('NeuronGroup'))

        Network_UI.Next_H_Block()

        tooltip_message = ''#str(self.variables).replace(',', '\r\n')
        for key in self.variables:
            tooltip_message+=str(self.variables[key])+' : '+str(key)+'\r\n'

        single_vars = list(self.variables.keys())
        mean_vars = ['np.mean(n.'+var+')' for var in self.variables]
        colors = list(self.variables.values())

        mean_curves = Network_UI.Add_plot_curve(y_label='NeuronGroup ' + str(single_vars), number_of_curves=len(mean_vars), names=mean_vars, colors=colors, lines=self.net_lines, tooltip_message=tooltip_message, return_list=True)
        self.net_curves=dict(zip(self.variables, mean_curves))

        Network_UI.Next_H_Block(stretch=0)

        self.neuron_title_label = Network_UI.Add_element(QLabel('Selected Neuron'))

        Network_UI.Next_H_Block()

        single_curves = Network_UI.Add_plot_curve(y_label='Neuron ' + str(single_vars), number_of_curves=len(single_vars), names=single_vars, colors=colors, lines=self.neuron_lines, tooltip_message=tooltip_message, return_list=True)
        self.neuron_curves = dict(zip(self.variables, single_curves))

        Network_UI.Next_H_Block(stretch=0)

        self.line_checkboxes = {}
        for var in self.variables:
            c=self.variables[var]
            txt='<font color='+('#%02x%02x%02x'%c).upper()+'>'+var+'</font>'
            Network_UI.Add_element(QLabel(txt))

        Network_UI.Next_H_Block(stretch=0)

        for var in self.variables:
            self.line_checkboxes[var] = Network_UI.Add_element(QCheckBox(''))
            self.line_checkboxes[var].setChecked(True)

    def update(self, Network_UI):
        if self.inh_exc_tab.isVisible():
            group = Network_UI.selected_neuron_group()

            self.group_title_label.setText('NeuronGroup ('+Network_UI.selected_neuron_group().tags[0]+')')

            for var in self.variables:
                iterations = group['n.iteration', 0, 'np'][-self.timesteps:]

                try:
                    if self.line_checkboxes[var].isChecked():
                        mean_var = 'np.mean(n.'+var+')'
                        data_mean = group[mean_var, 0, 'np'][-self.timesteps:]
                        self.net_curves[var].setData(iterations, data_mean)
                    else:
                        self.net_curves[var].clear()
                except:
                    self.net_curves[var].clear()
                    #print(var, "cannot be evaluated")

                try:
                    if self.line_checkboxes[var].isChecked():
                        single_var = 'n.' + var
                        data_neuron = group[single_var, 0, 'np'][-self.timesteps:, Network_UI.selected_neuron_id()].astype(def_dtype)
                        self.neuron_curves[var].setData(iterations, data_neuron)
                    else:
                        self.neuron_curves[var].clear()
                except:
                    self.neuron_curves[var].clear()
                    #print(var, "cannot be evaluated")




from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Visualization_Helper import *
import os


class code_execution_tab(TabBase):

    sample_code1='''import matplotlib.pyplot as plt

synapse_tag='GLU'
synapse_var='W'

combined = get_combined_syn_mats(n.afferent_synapses[synapse_tag], None,synapse_var)

for mat in combined.values():

	plt.imshow(mat, vmin=0, vmax=0.005)#cmap='gray', 
	plt.show()
'''

    sample_code2='''num=0

for syn in n.afferent_synapses['GLU']:

	num+=np.sum(syn.W>0.001)

print(num)
'''

    def initialize(self, Network_UI):
        self.Network_UI=Network_UI

        #Network_UI.Next_H_Block()

        #self.timed_execution_list_view = Network_UI.Add_element(QListView(), sidebar=True)
        #self.timed_execution_model=QtGui.QStandardItemModel()
        #self.timed_execution_list_view.setModel(self.timed_execution_model)

        #item = QtGui.QStandardItem("TEST")  # str(execution_time)+' | '+base_name)
        #item.setCheckable(True)
        #check = QtCore.Qt.Checked  # if checked else QtCore.Qt.Unchecked
        #item.setCheckState(check)
        #self.timed_execution_model.appendRow(item)

        self.code_execution_tab = Network_UI.Next_Tab('Code Exec.')
        self.comboBox = Network_UI.Add_element(QComboBox())

        self.txt_py_dict={}

        self.txt_py_dict[''] = ''
        self.txt_py_dict["show synapse matrices"] = self.sample_code1
        self.txt_py_dict["count synapses"] = self.sample_code2

        try:
            data_folder = get_data_folder(create_when_not_found=False)+'/scripts'

            if not os.path.exists(data_folder):
                os.mkdir(data_folder)

            for d in os.listdir(data_folder):
                if '.py' in d:
                    with open(data_folder+'/'+d) as f:
                        self.txt_py_dict[d.replace('.py','')] = f.read()

                    self.add_timed_script(d)
        except:
            pass

        for txt in self.txt_py_dict:
            self.comboBox.addItem(txt)


        def on_cb_click(event):
            txt=self.comboBox.itemText(self.comboBox.currentIndex())
            self.code_field.setText(self.txt_py_dict[txt])
            self.code_field.setStyleSheet("QTextEdit { background-color: rgb(255, 255, 255); }")

        def on_code_changed():
            self.code_field.setStyleSheet("QTextEdit { background-color: rgb(255, 210, 210); }")

        self.comboBox.currentIndexChanged.connect(on_cb_click)

        Network_UI.Next_H_Block()

        self.code_field=Network_UI.Add_element(QTextEdit())

        self.code_field.setAcceptRichText(False)

        self.code_field.textChanged.connect(on_code_changed)

        Network_UI.Next_H_Block(stretch=0.1)

        self.exec_btn=Network_UI.Add_element(QPushButton('Execute'))

        def exec_btn_click(event):
            net = Network_UI.network
            network = net
            n = Network_UI.selected_neuron_group()
            ng = n
            neurons = n
            neuron_group = n

            exec(self.code_field.toPlainText())

            print('code block executed successfuly!')

        self.exec_btn.clicked.connect(exec_btn_click)

        self.save_btn_click = Network_UI.Add_element(QPushButton('Save Script'))

        def save(input_txt):
            data_folder = get_data_folder(create_when_not_found=False)+'/scripts'
            name=data_folder + '/'+input_txt+'.py'
            with open(name, "w") as f:
                f.write(self.code_field.toPlainText())

            n=input_txt
            found = n in self.txt_py_dict

            self.txt_py_dict[n] = self.code_field.toPlainText()

            if not found:
                self.comboBox.addItem(n)
                self.comboBox.setCurrentIndex(len(self.txt_py_dict)-1)

            self.code_field.setStyleSheet("QTextEdit { background-color: rgb(255, 255, 255); }")

            #self.add_timed_script(name)

        def save_btn_click(event):
            Network_UI.text_input_dialog("Please name your script", 'save', save, self.comboBox.itemText(self.comboBox.currentIndex()).replace('.py',''))

        self.save_btn_click.clicked.connect(save_btn_click)

        self.compiled = None
        self.compiled_script_txt = ''
        self.timestep_execute_cb = Network_UI.Add_element(QCheckBox('Execute every timestep'))

    def update(self, Network_UI):

        if self.timestep_execute_cb.isChecked():

            net = Network_UI.network
            network = net
            n = Network_UI.selected_neuron_group()
            ng = n
            neurons = n
            neuron_group = n

            code = self.code_field.toPlainText()

            if self.compiled is None or self.compiled_script_txt != code:
                self.compiled = compile(code, '<string>', 'exec')
                self.compiled_script_txt = code

            exec(self.compiled)




from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Visualization_Helper import *

class isi_tab(TabBase):

    def __init__(self, title='Activity/ISI', param='output', mask_param='Input_Mask', timesteps=1000, mask_color_add=(-100, -100, -100)):
        super().__init__(title)
        self.param = param
        self.timesteps=timesteps
        self.mask_param = mask_param
        if self.mask_param is not None:
            self.compiled_param = compile('n.'+self.mask_param, '<string>', 'eval')
            self.inverted_compiled_param = compile('np.invert(n.' + self.mask_param+')', '<string>', 'eval')
            self.mask_color_add = mask_color_add

    def add_recorder_variables(self, neuron_group, Network_UI):
        if hasattr(neuron_group, self.param):
            Network_UI.add_recording_variable(neuron_group, 'n.' + self.param, timesteps=self.timesteps)
        if hasattr(neuron_group, self.param):
            self.sum_tag='np.sum(n.'+self.param+')'
            Network_UI.add_recording_variable(neuron_group, self.sum_tag, timesteps=1000)

    def initialize(self, Network_UI):
        self.isi_tab = Network_UI.Next_Tab(self.title)

        _, self.neuron_isi_plt = Network_UI.Add_plot_curve('neuron inter spike interval hist', True, False, legend=False, x_label='ISI', y_label='Frequency')
        _, self.net_isi_plt = Network_UI.Add_plot_curve('network inter spike interval hist', True, False, legend=False, x_label='ISI', y_label='Frequency')
        if self.mask_param is not None:
            _, self.input_isi_plt = Network_UI.Add_plot_curve('input inter spike interval hist', True, False, legend=False, x_label='ISI', y_label='Frequency')

        Network_UI.Next_H_Block()
        self.neuron_isi_bin_slider = QSlider(1)  # QtCore.Horizontal
        self.neuron_isi_bin_slider.setMinimum(1)
        self.neuron_isi_bin_slider.setMaximum(100)
        self.neuron_isi_bin_slider.setSliderPosition(10)
        self.neuron_isi_bin_slider.mouseReleaseEvent = Network_UI.static_update_func
        self.neuron_isi_bin_slider.setToolTip('slide to change bin count')
        Network_UI.Add_element(self.neuron_isi_bin_slider)  # , stretch=0.1

        self.net_isi_bin_slider = QSlider(1)  # QtCore.Horizontal
        self.net_isi_bin_slider.setMinimum(1)
        self.net_isi_bin_slider.setMaximum(100)
        self.net_isi_bin_slider.setSliderPosition(10)
        self.net_isi_bin_slider.mouseReleaseEvent = Network_UI.static_update_func
        self.net_isi_bin_slider.setToolTip('slide to change bin count')
        Network_UI.Add_element(self.net_isi_bin_slider)  # , stretch=0.1


        if self.mask_param is not None:
            self.input_isi_bin_slider = QSlider(1)  # QtCore.Horizontal
            self.input_isi_bin_slider.setMinimum(1)
            self.input_isi_bin_slider.setMaximum(100)
            self.input_isi_bin_slider.setSliderPosition(10)
            self.input_isi_bin_slider.mouseReleaseEvent = Network_UI.static_update_func
            self.input_isi_bin_slider.setToolTip('slide to change bin count')
            Network_UI.Add_element(self.input_isi_bin_slider)  # , stretch=0.1

        Network_UI.Next_H_Block()
        Network_UI.Add_element(QLabel(''))
        self.net_isi_cb = Network_UI.Add_element(QCheckBox('Compute network isi'))
        #self.net_isi_cb.setChecked(True)
        if self.mask_param is not None:
            self.input_isi_cb = Network_UI.Add_element(QCheckBox('Compute input isi'))
            #self.input_isi_cb.setChecked(True)

        Network_UI.Next_H_Block()
        self.neuron_avg_act_label = Network_UI.Add_element(QLabel('-'))
        _, self.net_avg_hist_plt = Network_UI.Add_plot_curve('network avg activities', True, False, legend=False, x_label='average activity', y_label='Frequency')
        if self.mask_param is not None:
            _, self.input_avg_hist_plt = Network_UI.Add_plot_curve('input avg activities', True, False, legend=False, x_label='average activity', y_label='Frequency')

        Network_UI.Next_H_Block()

        Network_UI.Add_element(QLabel(''))

        self.net_avg_bin_slider = QSlider(1)  # QtCore.Horizontal
        self.net_avg_bin_slider.setMinimum(1)
        self.net_avg_bin_slider.setMaximum(100)
        self.net_avg_bin_slider.setSliderPosition(10)
        self.net_avg_bin_slider.mouseReleaseEvent = Network_UI.static_update_func
        self.net_avg_bin_slider.setToolTip('slide to change bin count')
        Network_UI.Add_element(self.net_avg_bin_slider)  # , stretch=0.1

        if self.mask_param is not None:
            self.input_avg_bin_slider = QSlider(1)  # QtCore.Horizontal
            self.input_avg_bin_slider.setMinimum(1)
            self.input_avg_bin_slider.setMaximum(100)
            self.input_avg_bin_slider.setSliderPosition(10)
            self.input_avg_bin_slider.mouseReleaseEvent = Network_UI.static_update_func
            self.input_avg_bin_slider.setToolTip('slide to change bin count')
            Network_UI.Add_element(self.input_avg_bin_slider)  # , stretch=0.1

        Network_UI.Next_H_Block(stretch=0.1)

        Network_UI.Add_element(QLabel('interval length: '+str(self.timesteps)+' steps'))





    def update_ISI(self, Network_UI, group, input_mask, not_input_mask, net_color_input):

        neuron_bins = self.neuron_isi_bin_slider.sliderPosition()
        act_data = group['n.' + self.param, 0, 'np'][-self.timesteps:,:]

        self.neuron_isi_plt.clear()
        y, x = np.histogram(SpikeTrain_ISI(act_data[-self.timesteps:, Network_UI.selected_neuron_id()]), bins=neuron_bins)
        curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=Network_UI.neuron_select_color)
        self.neuron_isi_plt.addItem(curve)

        self.net_isi_plt.clear()
        if self.net_isi_cb.isChecked():
            if not_input_mask is True:
                net_act_data = act_data
            else:
                net_act_data=act_data[:,not_input_mask]
            net_bins = self.net_isi_bin_slider.sliderPosition()
            net_hist_data = []
            for i in range(net_act_data.shape[1]):
                net_hist_data+=SpikeTrain_ISI(net_act_data[:,i])
            y, x = np.histogram(net_hist_data, bins=net_bins)
            curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=group.color)
            self.net_isi_plt.addItem(curve)

        if self.mask_param is not None:
            self.input_isi_plt.clear()
            if input_mask is not False and self.input_isi_cb.isChecked():
                input_act_data = act_data[:, input_mask]
                input_bins = self.input_isi_bin_slider.sliderPosition()
                input_hist_data = []
                for i in range(input_act_data.shape[1]):
                    input_hist_data+=SpikeTrain_ISI(input_act_data[:,i])
                y, x = np.histogram(input_hist_data, bins=input_bins)
                curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=net_color_input)
                self.input_isi_plt.addItem(curve)

    def update_Mean_Activity(self, Network_UI, group, input_mask, not_input_mask, net_color_input):
        #rec = Network_UI.rec(group, self.timesteps)

        net_bins = self.net_avg_bin_slider.sliderPosition()

        avg_acts = np.mean(group['n.' + self.param, 0, 'np'][-self.timesteps:, :], axis=0)

        color_str=('#%02x%02x%02x'% Network_UI.neuron_select_color[0:3]).upper()
        self.neuron_avg_act_label.setText('Selected neuron average spike rate:<br><font color='+color_str+'>'+str(avg_acts[Network_UI.selected_neuron_id()])+'</font>')

        self.net_avg_hist_plt.clear()
        y, x = np.histogram(avg_acts[not_input_mask], bins=net_bins)
        curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=group.color)
        self.net_avg_hist_plt.addItem(curve)

        if self.mask_param is not None:
            self.input_avg_hist_plt.clear()
            if input_mask is not False:
                input_bins = self.input_avg_bin_slider.sliderPosition()
                y, x = np.histogram(avg_acts[input_mask], bins=input_bins)
                curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=net_color_input)  # todo:make faster!
                self.input_avg_hist_plt.addItem(curve)  # todo:make faster!

    def update(self, Network_UI):
        if self.isi_tab.isVisible():

            group = Network_UI.selected_neuron_group()
            n = group#for eval comand

            if self.mask_param is not None and hasattr(group, self.mask_param):
                input_mask = eval(self.compiled_param)
                not_input_mask = eval(self.inverted_compiled_param)
                mca = self.mask_color_add
            else:
                input_mask = False
                not_input_mask = True
                mca = (0, 0, 0)

            net_color_input = np.clip([group.color[0] + mca[0], group.color[1] + mca[1], group.color[2] + mca[2], 255], 0, 255)

            if hasattr(group, self.param):
                self.update_ISI(Network_UI, group, input_mask, not_input_mask, net_color_input)
                self.update_Mean_Activity(Network_UI, group, input_mask, not_input_mask, net_color_input)



from configparser import ConfigParser
import numpy as np
import zipfile
import time
from subprocess import call
import os
import pickle
import imageio


#import matplotlib.pylab as plt
#storage_manager_folder = '../../Data/StorageManager/'

def create_folder_if_not_exist(path):
    if not os.path.exists(path):
        os.mkdir(path)

def get_data_folder(create_when_not_found=True):
    path='./'
    while os.path.isdir(path):
        if os.path.isdir(path+'Data/'):
            return os.path.abspath(path+'Data/').replace('\\', '/')
        path='../'+path

    if not os.path.exists('./Data'):
        try:
            os.mkdir('./Data')
        except:
            print('was not able to create Data folder')
    return './Data'
    #raise Exception('No "Data" folder found above current working directory! Please create dirctory: ".../Project/Data".')

def zipDir(dirPath, zipPath, filter):
    zipf = zipfile.ZipFile(zipPath, mode='w')
    lenDirPath = len(dirPath)
    for root, _, files in os.walk(dirPath):
        for file in files:
            filter_file=False
            for f in filter:
                if f in root or f in file+'\\':
                    filter_file = True
            if not filter_file:
                filePath = os.path.join(root, file)
                print(filePath)
                zipf.write(filePath , filePath[lenDirPath :] )
    zipf.close()

class StorageManager:

    def dict_to_folder_name(self, dict):
        result = ''
        for k, v in dict.items():
            if result != '':
                result += '_'
            result += str(k).replace(' ', '')+'='+str(v).replace(' ', '')
        return result

    def __init__(self, main_folder_name, folder_name=None, random_nr=False, print_msg=True, add_new_when_exists=True, data_folder=get_data_folder(), use_evolution_path=True):

        if use_evolution_path:
            import PymoNNto.Exploration.Evolution.Interface_Functions as evo_func
            if evo_func.get_gene('evo_name', None) is not None and evo_func.get_gene('gen', None) is not None and evo_func.get_gene('id', None) is not None:
                main_folder_name = evo_func.get_gene('evo_name', None)
                folder_name = evo_func.get_gene_file(evo_func.evolution_genome)
                add_new_when_exists = False
                random_nr=False

        if type(main_folder_name) is dict:
            main_folder_name = self.dict_to_folder_name(main_folder_name)

        storage_manager_folder = data_folder +'/StorageManager/'

        if not os.path.exists(storage_manager_folder):
            try:
                os.mkdir(storage_manager_folder)
            except:
                print('SM folder already exists...')

        if folder_name is None:
            folder_name = main_folder_name

        self.folder_name = folder_name
        if type(random_nr) == bool and random_nr:
            self.folder_name += str(int(np.random.rand() * 10000))
        if type(random_nr) == int:
            self.folder_name += str(random_nr)

        if not os.path.exists(storage_manager_folder+main_folder_name+'/'):
            try:
                os.mkdir(storage_manager_folder+main_folder_name+'/')
            except:
                print("main folder already exists")

        if add_new_when_exists:
            if os.path.isdir(storage_manager_folder+main_folder_name+'/'+self.folder_name+'/'):
                count = 0
                while os.path.isdir(storage_manager_folder+main_folder_name+'/'+self.folder_name+'_{}/'.format(count)):
                    count += 1
                self.folder_name += '_{}'.format(count)

        #self.foldername += '_'+count#int(np.random.rand() * 10*add_randum_number)

        self.absolute_path = storage_manager_folder+main_folder_name+'/'+self.folder_name+'/'

        new = False
        if not os.path.exists(self.absolute_path):
            try:
                os.mkdir(self.absolute_path)
            except:
                print('target folder already exits')
            new = True

        self.config_file_name = 'config.ini'
        self.config=None

        self.frame_counter = {}

        if new:
            if print_msg:
                print(self.absolute_path)
            self.save_param('time', time.time())

    def init_config(self):
        if self.config is None:
            self.config = ConfigParser()
            self.config.read(self.absolute_path + self.config_file_name)

    def save_recorder(self, tag, recorder, keys=[]):
        if type(keys) is str:
            keys=[keys]
        if len(keys) == 0:
            keys=recorder.variables.keys()
        for key in keys:
            self.save_np(tag+key, np.array(recorder.variables[key]))

    def save_np(self, key, obj):
        np.save(self.absolute_path + key + '.npy', arr=obj)

    def save_obj(self, key, obj):
        pickle.dump(obj, open(self.absolute_path + key + '.obj', 'wb'))

    def load_obj(self, key):
        return pickle.load(open(self.absolute_path + key + '.obj', 'rb'))

    def has_obj(self, key):
        return os.path.isfile(self.absolute_path + key + '.obj')

    def save_param_dict(self, d, section='Parameters'):
        for key in d:
            self.save_param(key, d[key], section)

    def save_param(self, key, value, section='Parameters'):
        self.init_config()
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, key, str(value))
        with open(self.absolute_path+self.config_file_name, 'w') as configfile:
            self.config.write(configfile)

    def get_next_frame_name(self, key):
        if not key in self.frame_counter:
            self.frame_counter[key] = 0
        self.frame_counter[key] += 1
        return self.absolute_path+key+'{}.png'.format(self.frame_counter[key]-1)


    def save_frame(self, image, key):
        #plt.imshow(image)  # Needs to be in row,col order
        #plt.savefig(self.absolute_path+key+'{}.png'.format(self.frame_counter[key]))
        #im = Image.fromarray(image)
        #im.save(self.absolute_path+key+'{}.png'.format(self.frame_counter[key]))
        imageio.imwrite(self.get_next_frame_name(key), (image).astype(np.uint8))#
        #scipy.misc.imsave(self.absolute_path+key+'{}.png'.format(self.frame_counter[key]), image)



    def render_video(self, key, delete_images, reset_frame_counter=True):

        #ffmpeg -r 25 -i frame%09d.png -vcodec mpeg4 -b 10000000 -y video.mp4
        #call(['ffmpeg', '-r', '25', '-i', self.absolute_path + key + '%d.png', '-vcodec', 'mpeg4', '-q:v', '1', '-b', '10000000', '-y', self.absolute_path + key + '.mp4'], shell=True)#

        # ffmpeg -i temp-%d.png -c:v libx264 -strict -2 -preset slow -pix_fmt yuv420p -vf "scale=trunc(iw/2)*2:trunc(ih/2)*2" -f mp4 output.mp4 #, '-r', '25'
        call(['ffmpeg', '-i', self.absolute_path + key + '%d.png', '-c:v', 'libx264', '-strict', '-2', '-preset', 'slow', '-pix_fmt', 'yuv420p', '-vf', 'scale=trunc(iw/2)*2:trunc(ih/2)*2', '-f', 'mp4', self.absolute_path + key + '.mp4'], shell=True)
        if delete_images:
            files = os.listdir(self.absolute_path)
            for f in files:
                if key in f and '.png' in f:
                    #print(self.absolute_path+f)
                    #os.remove(f)#todo savety check
                    os.remove(self.absolute_path + f)

        if reset_frame_counter:
            self.frame_counter[key] = 0

    def load_param(self, key, section='Parameters', default=None, return_string=False):#casting!!!
        self.init_config()
        try:
            s = self.config.get(section, key)
        except:
            return default
        if return_string:
            return s
        if len(s) > 0:
            if s=='True':
                return True
            if s=='False':
                return False
            try:
                if '.' in s:
                    return float(s)
                else:
                    return int(s)
            except:
                return s
        return s


    def load_np(self, key):
        return np.load(self.absolute_path + key + '.npy')

    def copy_project_files(self, base_folder='../../'):
        zf = zipfile.ZipFile(self.absolute_path+"backup.zip", "w")
        for dirname, subdirs, files in os.walk(base_folder):#
            #print(dirname, subdirs, files)
            if dirname != base_folder and not 'Data' in dirname and not '.git' in dirname and not '.idea' in dirname and not '.zip' in dirname:
                zf.write(dirname)
                for filename in files:
                    zf.write(os.path.join(dirname, filename))
        zf.close()

class SimpleStorageManager(StorageManager):

    def __init__(self, folder):
        self.absolute_path = folder

        self.config_file_name = 'config.ini'
        self.config = None

        self.frame_counter = {}


class StorageManagerGroup:

    def __init__(self, Tag, main_folder_name=None, data_folder=get_data_folder()):

        storage_manager_folder = data_folder + '/StorageManager/'

        if main_folder_name is None:
            main_folder_name = Tag

        self.absolute_path = storage_manager_folder + main_folder_name + '/'

        self.StorageManagerList = []
        self.StorageManager_file_appendixes = []

        ct = 0
        for folder in os.listdir(self.absolute_path):
            if os.path.isdir(self.absolute_path+folder) and Tag in folder:
                self.StorageManagerList.append(StorageManager(Tag, folder, add_new_when_exists=False, data_folder=data_folder))

                #app = folder.split('_')[1:]
                #if len(app) > 1:
                #    self.StorageManagerList[-1].save_param('appendix', app[1])

                #self.StorageManagerList[-1].save_param('folder_number', ct)
                #ct += 1

    def sort_by(self, param, section='Parameters'):
        self.StorageManagerList = sorted(self.StorageManagerList, key=lambda sm: sm.load_param(param, section, default=-np.inf))

    def remove_None(self, l, remove_None=True):
        if remove_None:
            return list(filter(lambda a: a is not None, l))
        else:
            return l

    def get_param_list(self, param, section='Parameters', remove_None=False):
        return self.remove_None([sm.load_param(param, section) for sm in self.StorageManagerList], remove_None)

    def get_np_list(self, np_name, remove_None=False):
        return self.remove_None([sm.load_np(np_name) for sm in self.StorageManagerList], remove_None)

    def get_obj_list(self, obj_name, remove_None=False):
        return self.remove_None([sm.load_obj(obj_name) for sm in self.StorageManagerList], remove_None)

    def get_multi_param_list(self, params, section='Parameters', remove_None=True):
        remove=True
        results = []
        for param in params:
            results.append(np.array(self.get_param_list(param, section)))
            remove*=(results[-1]!=None)*(results[-1]!=np.nan)
        results = np.array(results)
        if remove_None:
            results=results[:,remove]#np.where(results is not None)#.any(axis=1)
        return results.astype(np.float64)

    def remove_duplicates_get_eval(self, x, y, evalstr='np.average(a)'):
        unique = np.unique(x)
        result = np.zeros(len(unique))
        for i, u in enumerate(unique):
            a = y[x == u]
            result[i] = eval(evalstr)
        return unique, result

    def __getitem__(self, query):
        result = []
        parts = query.replace('=', '|').replace(' ', '|').replace('>', '|').replace('<', '|').split('|')
        key_q = parts[0]

        for sm in self.StorageManagerList:
            param = sm.load_param(key_q, return_string=True)
            if param is not None:
                new_q = query.replace(key_q, param)
                if eval(new_q):
                    result.append(sm)

        return result







from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Reconstruct_Analyze_Label.Reconstruct_Analyze_Label import *

class sidebar_image_module(TabBase):

    def add_recorder_variables(self, neuron_group, Network_UI):
        return

    def initialize(self, Network_UI):

        if Network_UI.network['image_act', 0] is not None:

            def image_activator_on_off(event):
                Network_UI.network['image_act', 0].behaviour_enabled = self.input_select_box.currentText() != 'None'

            self.input_select_box = QComboBox()
            self.input_select_box.addItem("Images")
            self.input_select_box.addItem("None")
            self.input_select_box.currentIndexChanged.connect(image_activator_on_off)
            Network_UI.Add_Sidebar_Element(self.input_select_box)

            #self.image_history_image = Network_UI.Add_Image_Item(False, True, title='', stretch=0.3)
            self.image_history = []

            self.prediction_history_image = Network_UI.Add_Image_Item(False, True, title='', stretch=0.3)
            self.prediction_history = []

            self.backprop_image = Network_UI.Add_Image_Item(False, True, title='', stretch=0.3)

    def max_div(self, pattern):
        ms = np.max(pattern)
        if ms > 0:
            return pattern / ms
        else:
            return pattern.copy()

    def update(self, Network_UI):

        # save data timestep
        if not Network_UI.update_without_state_change and Network_UI.network['image_act', 0] is not None:
            image_act = Network_UI.network['image_act', 0]

            #index = image_act.current_pattern_index
            #if index >- 1:
            #    pattern = image_act.patterns[image_act.current_pattern_index].copy()
            #    pattern.resize((3, image_act.grid_width, image_act.grid_height))
            #    self.image_history.append(np.rot90(np.dstack(pattern), 3))
            #    self.image_history.append(np.ones((1, self.image_history[0].shape[1], 3)))
            #    self.image_history_image.setImage(np.concatenate(self.image_history, axis=0), levels=(0, 1))

            group = Network_UI.selected_neuron_group()
            if hasattr(group, 'Input_Mask'):
                pattern = self.max_div(group.output[group.Input_Mask])
                #print(pattern)
                pattern.resize((3, image_act.grid_height, image_act.grid_width))
                #group['n.output', 0, 'np'][]
                self.prediction_history.append(np.rot90(np.dstack(pattern), 3))
                self.prediction_history.append(np.ones((1, self.prediction_history[0].shape[1], 3)))
                self.prediction_history_image.setImage(np.concatenate(self.prediction_history, axis=0), levels=(0, 1))

                RALN = Reconstruct_Analyze_Label_Network(Network_UI.network)
                RALN.zero_recon()
                group.recon = group.output.copy()
                RALN.propagation('W', 6, 'backward', 'forget', 'all', temporal_recon_groups=[Network_UI.selected_neuron_group()], exponent=None, normalize=False, filter_weakest_percent=None)

                baseline = self.max_div(group.temporal_recon[-1][group.Input_Mask])
                images = []
                for r in group.temporal_recon:
                    pattern = np.clip(self.max_div(r[group.Input_Mask])-baseline/2, 0, None)
                    pattern.resize((3, image_act.grid_height, image_act.grid_width))
                    images.insert(0, np.rot90(np.dstack(pattern), 3))
                    images.insert(0, np.ones((1,  images[0].shape[1], 3)))
                self.backprop_image.setImage(np.concatenate(images, axis=0), levels=(0, 1))

            for _ in range(len(self.image_history)-5*2): self.image_history.pop(0)
            for _ in range(len(self.prediction_history)-5*2): self.prediction_history.pop(0)







from PymoNNto.Exploration.Evolution.Interface_Functions import *

score = get_gene('a', 0) + get_gene('b', 0) + get_gene('c', 0) + get_gene('d', 0)

set_score(score)



#print('genes: ', get_genome())
#print('gene a: ', get_gene('a', None))
#print('gene b: ', gene('b', 1))


import matplotlib
matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt

from PymoNNto.NetworkBehaviour.Input.Activator import *
from PymoNNto.NetworkCore.Neuron_Group import *

import PymoNNto.Exploration.Evolution_Old.NetworkEvaluationFunctions as EvalF



class visualization_recorder(Behaviour):
    def __init__(self):
        self.recorderlist = []
        self.neurons = None

    def set_variables(self, neurons):
        self.neurons=neurons
        for recorder in self.recorderlist:
            recorder.set_variables(neurons)

    def new_iteration(self, neurons):
        self.neurons = neurons
        for recorder in self.recorderlist:
            recorder.new_iteration(neurons)

class score_visualization_recorder(visualization_recorder):

    def __init__(self):
        super().__init__()
        self.neu_rec = Recorder(['n.norm_value', 'n.output_activity_history[0]', 'n.activity'], gapwidth=100)
        self.recorderlist.append(self.neu_rec)
        self.syn_rec = SynapseGroupRecorder(['syn.GLU_Synapses'], gapwidth=1000)
        self.recorderlist.append(self.syn_rec)

    def visualize(self):
        print('rendering score visualization:')

        fig, axes = plt.subplots(2, 3)
        fig.canvas.set_window_title('score visualization')

        #image = get_whole_Network_weight_image(self.neurons, neuron_src_groups=None)
        #axes[0, 2].imshow(image, interpolation="nearest")
        #axes[0, 2].set_xlabel('weights')

        print('- activity diversity score...')
        axes[0, 0].plot(EvalF.get_diversity_score_sliding_window(np.array(self.neu_rec['n.activity'])))
        print('- activity sparseness score...')
        axes[0, 1].plot(EvalF.get_sparseness_score_sliding_window(np.array(self.neu_rec['n.activity'])))
        print('- weight diversity score...')
        axes[1, 0].plot(EvalF.get_diversity_score_timeline(np.array(self.syn_rec['syn.GLU_Synapses']), self.neurons.get_combined_synapse_shape('GLU')))#(50,100)(400, 100)
        #axes[1, 1].get_TF_ANN_trining_accuracy_score

class activity_visualization_recorder(visualization_recorder):

    def __init__(self):
        super().__init__()
        self.neu_rec = Recorder(['n[8].avg', 'n.norm_value', 'n.output_activity_history[0]', 'n.activity'], gapwidth=100)
        self.recorderlist.append(self.neu_rec)

    def visualize(self):
        print('rendering activity visualization...')

        fig, axes = plt.subplots(2, 3)
        fig.canvas.set_window_title('activity visualization')

        axes[0, 0].plot(self.neu_rec['n[8].avg'],linewidth=2)
        axes[0, 0].axhline(y=np.average(self.neurons[HomeostaticMechanism].min))
        axes[0, 0].axhline(y=np.average(self.neurons[HomeostaticMechanism].max))
        axes[0, 0].set_xlabel('avg_act')

        axes[1, 0].plot(self.neu_rec['n.norm_value'],linewidth=2)
        axes[1, 0].set_xlabel('norm_value')

        axes[0, 1].plot(self.neu_rec['n.output_activity_history[0]'],linewidth=2)
        axes[0, 1].set_xlabel('norm_value')
        axes[1, 1].plot(self.neu_rec['n.activity'],linewidth=2)
        axes[1, 1].set_xlabel('activity')

        #axes[0, 2].plot(EvalF.get_diversity_score_sliding_window(np.array(self.neu_rec['n.activity'])))
        #axes[1, 2].plot(EvalF.get_sparseness_score_sliding_window(np.array(self.neu_rec['n.activity'])))

        # axes[0, 3].plot(np.array(syn_rec2['syn.GLU_Synapses'])[:, 0:20])#syn_rec.get_block(0, syn_rec.GLU_Synapses)

class input_visualization_recorder(visualization_recorder):

    def __init__(self):#todo: LGN n[8]
        super().__init__()
        self.neu_rec = Recorder(['n[8].avg', 'n.norm_value', 'n.output_activity_history[0]', 'n.activity'], gapwidth=100)
        self.recorderlist.append(self.neu_rec)

    def visualize(self):
        print('rendering input visualization...')

        fig, axes = plt.subplots(2, 4)
        fig.canvas.set_window_title('input visualization')

        axes[0, 0].plot(self.neu_rec['n[8].avg'])
        axes[0, 0].axhline(y=self.neurons[HomeostaticMechanism].min)
        axes[0, 0].axhline(y=self.neurons[HomeostaticMechanism].max)
        axes[0, 0].set_xlabel('avg_act')

        axes[1, 1].plot(self.neu_rec['n.norm_value'])
        axes[1, 1].set_xlabel('norm_value')

        axes[0, 1].plot(self.neu_rec['n.output_activity_history[0]'])
        axes[0, 1].set_xlabel('output_activity_history')
        axes[1, 1].plot(self.neu_rec['n.activity'])
        axes[1, 1].set_xlabel('activity')

        axes[0, 2].imshow(get_reconstruction_activations(np.array(self.neu_rec['n.activity'])[-10:-1], 10, 10), cmap=plt.cm.gist_gray, interpolation='nearest')

        axes[1, 2].matshow(np.array(self.neu_rec['n.norm_value'])[-1].reshape((10, 10)))
        axes[1, 2].set_xlabel('last norm_values')


        axes[0, 3].hist(np.array(self.neu_rec['n.activity'])[:, 0], bins=50)
        axes[0, 3].set_xlabel('first neuron activity distribution')

        axes[1, 3].hist(np.array(self.neu_rec['n.activity'])[:, 25], bins=50)
        axes[1, 3].set_xlabel('25. neuron activity distribution')

class weight_visualization_recorder(visualization_recorder):

    def __init__(self):
        super().__init__()
        self.syn_rec = SynapseGroupRecorder(['syn.GLU_Synapses'], gapwidth=100)
        self.recorderlist.append(self.syn_rec)

    def visualize(self):
        print('rendering weight visualization...')

        #fig, axes = plt.subplots(2, 3)
        fig = plt.figure()
        fig.canvas.set_window_title('weight visualization')

        syns = self.neurons.afferent_synapses.get(self.syn_rec.transmitter,[])
        input_size = syns[0].get_src_size()

        data = np.array(self.syn_rec['syn.GLU_Synapses'])

        fig.add_subplot(2, 3, 1).plot(data[:, 0:input_size])
        fig.add_subplot(2, 3, 2).matshow(data[:, 0:input_size].transpose())

        ax=fig.add_subplot(2, 3, 4)
        print(input_size, data.shape)
        ax.hist(data[0, 0:input_size], alpha=0.5, bins=50)
        ax.hist(data[-1, 0:input_size], alpha=0.5, bins=50)

        ax = fig.add_subplot(2, 3, 6, projection='3d')
        plot_3D_histogram(ax, data, 40, 10, x_label_caption='weights')

        #print(data.shape)
        while data.shape[0] < data.shape[1] and data.shape[1]%2 == 0:
            #x = data.shape[0]
            y = data.shape[1]

            data = np.concatenate((data[:, 0:int(y/2)], np.ones((5, int(y/2))), data[:, int(y/2):y]), axis=0)

        fig.add_subplot(2, 3, 3).matshow(data.transpose())




class firing_rate_visualization_recorder(visualization_recorder):

    def __init__(self):
        super().__init__()
        self.neu_rec = Recorder(['n.output_activity_history[0]', 'n.pre_inhibition_act', 'n.activity', 'n.norm_value'], gapwidth=1)
        self.recorderlist.append(self.neu_rec)

    def visualize(self):
        print('rendering firing rate visualization...')

        fig = plt.figure()
        fig.canvas.set_window_title('firing rate visualization')

        resolution = 100

        act = np.array(self.neu_rec['n.output_activity_history[0]'])[:, 0]
        act = act[act > 0]
        #plot_histogram(fig.add_subplot(2, 2, 1), act, resolution)
        plot_3D_histogram(fig.add_subplot(2, 2, 1, projection='3d'), act, resolution, 10, x_label_caption='output_activity_history')

        act = np.array(self.neu_rec['n.pre_inhibition_act'])[:, 0]
        act = act[act > 0]
        #plot_histogram(fig.add_subplot(2, 2, 2), np.array(self.neu_rec['n.pre_inhibition_act'])[:, 0], resolution)
        plot_3D_histogram(fig.add_subplot(2, 2, 2, projection='3d'), act, resolution, 10, x_label_caption='pre_inhibition_act')

        act = np.array(self.neu_rec['n.activity'])[:, 0]
        act = act[act > 0]
        #plot_histogram(fig.add_subplot(2, 2, 3), np.array(self.neu_rec['n.activity'])[:, 0], resolution)
        plot_3D_histogram(fig.add_subplot(2, 2, 3, projection='3d'), act, resolution, 10, x_label_caption='activity')

        fig.add_subplot(2, 2, 4).plot(self.neu_rec['n.norm_value'])
        #plot_3D_histogram(fig.add_subplot(2, 2, 4, projection='3d'), np.array(self.neu_rec['n.pre_inhibition_act'])[:, 0], resolution, 10, x_label_caption='pre_inhibition_act')


def visualize_layers(network, Cortex_PC_Neuron_Groups, LGN_PC_Neurons, v_split=False):
    num_groups = len(Cortex_PC_Neuron_Groups)
    fig, axes = plt.subplots(1, num_groups+(len(Cortex_PC_Neuron_Groups)==1))#int(num_groups/2), int(num_groups-num_groups/2))
    for i, group in enumerate(Cortex_PC_Neuron_Groups):
        #axes[i].imshow(get_group_block_mat_reconstruction_image(network, block, 'GLU_Synapses',3, individual_norm=True))

        image = EvalF.sum_cycles(EvalF.get_reconstruction(network, group))

        #image = EvalF.sum_temporal_abstraction_steps(network.get_reconstruction(group, None, LGN_PC_Neurons, 'GLU_Synapses', group.reconstruction_steps, individual_norm=True))
        axes[i].imshow(get_RGB_neuron_weight_image([image, None, None], [group.height*group.depth, group.width], [LGN_PC_Neurons.height*LGN_PC_Neurons.depth, LGN_PC_Neurons.width], v_split_first=v_split))#pattern_f
    plt.tight_layout()

def visualize_input_and_learned_patterns(network, Cortex_PC_Neurons, LGN_PC_Neurons, pattern=None, NN_eval=False, v_split=False):#, split_green_red_overlay=False
    print('rendering input and learned patterns visualization...')

    fig, axes = plt.subplots(2, 2)
    fig.canvas.set_window_title('input and learned patterns')

    image = get_whole_Network_weight_image(Cortex_PC_Neurons, neuron_src_groups=None, individual_norm=True)
    axes[1, 1].imshow(image, interpolation="nearest")

    #print('reconstruction')
    image = network.get_reconstruction(Cortex_PC_Neurons, list(range(Cortex_PC_Neurons.size)), LGN_PC_Neurons, 'GLU_Synapses', 3, individual_norm=True)
    #axes[0, 1].imshow(get_reconstruction_activations(image, 10, 10), cmap=plt.cm.gist_gray, interpolation='nearest')#, interpolation="nearest")

    axes[0, 1].imshow(get_RGB_neuron_weight_image([image, None, None], [Cortex_PC_Neurons.height*Cortex_PC_Neurons.depth, Cortex_PC_Neurons.width], [LGN_PC_Neurons.height*LGN_PC_Neurons.depth, LGN_PC_Neurons.width], v_split_first=v_split))#pattern_f

    #axes[0, 1].imshow(get_combined_RGB_Image(image, None, Cortex_PC_Neurons.width, Cortex_PC_Neurons.height*Cortex_PC_Neurons.depth, LGN_PC_Neurons.width, LGN_PC_Neurons.height*LGN_PC_Neurons.depth, transform_mat_func=pattern_f), interpolation='nearest')  # , interpolation="nearest")
    axes[0, 1].set_xlabel('weights')

    #print('reconstruction act')
    axes[0, 0].imshow(get_reconstruction_activations(np.clip(np.array(LGN_PC_Neurons[NeuronActivator].get_pattern_samples(10)), 0, 1), LGN_PC_Neurons.width, LGN_PC_Neurons.height * LGN_PC_Neurons.depth), cmap=plt.cm.gist_gray, interpolation='nearest')
    axes[0, 0].set_xlabel('input')

    if NN_eval:
        for i in range(3):
            axes[1, 1].plot(EvalF.get_TF_ANN_training_accuracy_score(network, [Cortex_PC_Neurons], [LGN_PC_Neurons[NeuronActivator].TNAPatterns[0]], 1000, 1000))

    #if not split_green_red_overlay:
    #    big_img = EvalF.get_pattern_response_image(network, LGN_PC_Neurons[TRENNeuronActivator].TNAPatterns[0], LGN_PC_Neurons, [Cortex_PC_Neurons])

        #big_img = EvalF.get_sorted_overview_image(network, [Cortex_PC_Neurons], [LGN_PC_Neurons[TRENNeuronActivator].TNAPatterns[0]],100, LGN=LGN_PC_Neurons)

        #axes[1, 0].matshow(big_img, cmap=plt.cm.gray)
        #axes[1, 0].set_xlabel('network resopnses')


def plot_network(network, blocks=None):
    fig = plt.figure()
    ax = fig.add_subplot(111)#, projection='3d'

    arrow_width=1
    arrow_lenth=1

    #img = mpimg.imread('../Data/Img/5.png')
    #ax.imshow(img)

    #x=[20, 100, 100, 110, 110, 100, 100, 80, 80, 60]
    #y=[0,  0,   10,  10,  30,  30,  40,  40, 45, 45]
    #z=[0,  0,   0,   0,   0,   0,   0,   0,  0,  0]
    #plt.plot(x,y,z)

    def pol2cart(rho, phi):
        x = rho * np.cos(phi)
        y = rho * np.sin(phi)
        return x, y

    for ng in network.NeuronGroups:
        ax.scatter(ng.x, ng.y)#, ng.z
        ax.text(np.min(ng.x), np.average(ng.y), ng.tags[0], color="blue")

    if blocks is not None:
        for ng in blocks:
            ax.scatter(ng.x, ng.y)#, ng.z
            ax.text(np.min(ng.x), np.average(ng.y), ng.tags[0])

    for sg in network.SynapseGroups:
        mul = 1
        color = 'green'
        if 'GABA' in sg.tags:#hasattr(sg, 'GABA_Synapses'):
            color = 'red'
            mul = -1

        if sg.src is not sg.dst:
            a_x = np.average(sg.src.x)+mul
            a_y = np.average(sg.src.y)
            d_x = np.average(sg.dst.x)+mul
            d_y = np.average(sg.dst.y)

            plt.plot([a_x, d_x], [a_y, d_y], color=color)#[np.average(sg.src.z), np.average(sg.dst.z)]
            plt.arrow(a_x, a_y, d_x-a_x, d_y-a_y, color=color, head_width=arrow_width, head_length=arrow_lenth)
        else:

            vec = np.array([sg.src.x, sg.src.y, sg.src.z])
            av = np.average(vec, axis=1)

            m = np.max(vec, axis=1)

            if 'GABA' in sg.tags:
                m[1] = np.min(vec[1])

            center = (av + m) / 2

            dist = np.linalg.norm(center-m, ord=2)

            xps = []
            yps = []
            zps = []

            for i in range(140):
                x, y = pol2cart(dist*0.7, np.deg2rad((i-10)*mul))
                xps.append(x+center[0])
                yps.append(y+center[1])
                zps.append(center[2])

            plt.plot(xps, yps, color=color)#, z
            plt.arrow(xps[-2], yps[-2], xps[-1]-xps[-2], yps[-1]-yps[-2], color=color, head_width=arrow_width, head_length=arrow_lenth)


    ax.set_xlabel('X Label')
    ax.set_ylabel('Y Label')
    #ax.set_zlabel('Z Label')

    #ax.set_xlim(-5, 5)
    #ax.set_ylim(-2, 12)
    #ax.set_zlim(0, 100)

    plt.show()

def plot(data):
    plt.plot(data)

def matshow(data):
    plt.matshow(data)

def imshow(data):
    plt.imshow(data, cmap=plt.cm.gist_gray, interpolation='nearest')

def run_and_visualize(behaviours, network, Cortex_PC_Neurons, iterations):
    network.add_behaviours_to_neuron_group(behaviours, Cortex_PC_Neurons)
    network.simulate_iterations(100, iterations / 100, True)
    for b in behaviours:
        b.visualize()


def show():
    plt.show()



from PymoNNto.Exploration.Evolution.Evolution import *

if __name__ == '__main__':
    genome = {'a': 1, 'b': 2, 'c': 2, 'd': 2, 'e':3}

    evo = Evolution(name='my_test_evo',
                    slave_file='test_slave.py', #Exploration/Evolution/
                    individual_count=10,
                    mutation=0.04,
                    death_rate=0.5,
                    constraints=['b>=a', 'a<=1', 'b>= 2'],
                    inactive_genome_info={'info': 'my_info'},
                    start_genomes=[genome],
                    devices={'single_thread': 0,
                             'multi_thread': 4,
                             'ssh vieth@poppy.fias.uni-frankfurt.de': 0,
                             'ssh marius@hey3kmuagjunsk2b.myfritz.net': 0,
                             }
                    )

    if not evo.start(ui=True):
        evo.continue_evolution(ui=True)



from PymoNNto.Exploration.StorageManager.StorageManager import *
import paramiko
from scp import SCPClient

def is_invalid_evo_name(name):
    return '/' in name or '.' in name or ' ' in name or '   ' in name or '\\' in name or name in ['Documents', 'Pictures', 'Music', 'Public', 'Videos', 'Dokumente', 'Bilder', 'Musik', 'Downloads', 'Öffentlich']


def get_ssh_connection(host, user, password):
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(host, username=user, password=password)
    return ssh

def get_response(out, err):
    result = []
    for line in out.read().splitlines():
        result.append(line.decode("utf-8"))
    for line in err.read().splitlines():
        result.append(line.decode("utf-8"))
    return result

def split_ssh_user_host_password_string(user_host_pw_str):
    password = None
    user = None
    host = None
    user_host = user_host_pw_str.replace('ssh ', '').split('@')
    if len(user_host) == 2:
        user = user_host[0]
        host = user_host[1]
        if ' ' in host:
            host_pw = host.split(' ')
            if len(host_pw) == 2:
                host = host_pw[0]
                password = host_pw[1]
                print('warning: a clear text ssh password inside of the code is dangerous!')
            else:
                print('space in host but no password detected')
    else:
        print('cannot split user and host')

    return user, host, password

def zip_project(Main_Folder, zip_file_path):
    zipDir(Main_Folder, zip_file_path, ['.git', '.zip', '.idea', '\\Evolution_Project_Clones\\', '\\StorageManager\\', '\\NetworkStates\\', '\\Evo\\', '\\__pycache__\\', '\\midis\\'])

def get_epc_folder(create=True):
    Data_Folder = get_data_folder()
    if Data_Folder != './Data':
        epc_folder = Data_Folder + '/Evolution_Project_Clones'
        create_folder_if_not_exist(epc_folder)
        return epc_folder
    return None

def clone_project(name):
    epc_folder = get_epc_folder()
    if epc_folder is not None:
        Main_Folder = epc_folder.replace('Data/Evolution_Project_Clones', '')

        src = epc_folder + '/' + name + '.zip'
        zip_project(Main_Folder, src)

        zipf = zipfile.ZipFile(src, mode='r')
        zipf.extractall(path=epc_folder + '/' + name, members=None, pwd=None)
        zipf.close()

        os.remove(src)
        return epc_folder+'/'+name+'/'
    else:
        print('Error No root "Data" folder found')

def transfer_project(name, user, host, password=None):
    # search main project folder
    Data_Folder = get_data_folder()
    if Data_Folder != './Data':
        Main_Folder = Data_Folder.replace('Data', '')

        src = Data_Folder + '/' + name + '.zip'
        dst = name + '.zip'

        # zip project
        print('Create Zip File from current folder:')
        zip_project(Main_Folder, src)

        # transfer
        print('Transfer zip to', host)
        ssh = get_ssh_connection(host, user, password)
        scp = SCPClient(ssh.get_transport())
        scp.put(src, dst)

        # remove zip
        os.remove(src)

        # remote unzip and remote remove zip
        print('Extract zip at', host)
        cmd = 'rm -r -d ' + name + '; '  # remove old directory
        cmd += 'mkdir ' + name + '; '  # create new directory
        cmd += 'unzip ' + name + '.zip -d ' + name + '; '  # unzip
        cmd += 'rm ' + name + '.zip'  # remove zip
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
        print(get_response(ssh_stdout, ssh_stderr))

        scp.close()
        ssh.close()
    else:
        print('Error No root "Data" folder found')

def get_Data(name, user, host, password):
    # search main project folder
    Data_Folder = get_data_folder()
    if Data_Folder != './Data':

        src = name + '/Data.zip'
        dst_path = get_epc_folder()+'/'+name+'/' #test.zip
        dst = dst_path + 'Data.zip'

        print(src,dst_path,dst)

        ssh = get_ssh_connection(host, user, password)

        # zip project
        cmd = 'cd ' + name + ' ; '
        cmd += 'zip -r Data.zip Data ;'
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
        print(get_response(ssh_stdout, ssh_stderr))

        # transfer
        print('Transfer zip from', host)
        scp = SCPClient(ssh.get_transport())
        scp.get(src, dst)

        with zipfile.ZipFile(dst, "r") as zip_ref:
            zip_ref.extractall(dst_path)

        # remove zip
        cmd = 'cd ' + name + '; '
        cmd += 'rm Data.zip; '
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(cmd)
        print(get_response(ssh_stdout, ssh_stderr))
        os.remove(dst)

        scp.close()
        ssh.close()
    else:
        print('Error No root "Data" folder found')

def ssh_execute_evo(server, name, python_cmd='python3'):
    user, host, password = split_ssh_user_host_password_string(server)
    ssh = get_ssh_connection(host, user, password)

    command = 'cd ' + name + '; '
    #command = 'nano .bashrc'
    command += 'screen -dmS ' + name + ' sh; screen -S ' + name + ' -X stuff "'+python_cmd+' execute_evolution.py \r\n"'

    ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command)
    response = get_response(ssh_stdout, ssh_stderr)
    print(response)

    ssh.close()

def ssh_stop_evo(server, name, remove_evo=False):
    user, host, password = split_ssh_user_host_password_string(server)
    ssh = get_ssh_connection(host, user, password)

    command = 'screen -XS ' + name + ' quit;'
    ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command)
    response = get_response(ssh_stdout, ssh_stderr)
    print(response)

    if remove_evo:
        command = 'rm -r ' + name
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command)
        response = get_response(ssh_stdout, ssh_stderr)
        print(response)

    ssh.close()

def ssh_get_running(server):
    response=''
    try:
        user, host, password = split_ssh_user_host_password_string(server)
        ssh = get_ssh_connection(host, user, password)

        command = 'screen -ls'
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command, timeout=10)
        response = get_response(ssh_stdout, ssh_stderr)
        print(response)

        ssh.close()
    except:
        print(server, 'no response')

    result = []

    for l in response:
        if '	' in l:
            dot_split = l.split('.')
            if len(dot_split) > 1:
                space_split = dot_split[1].split('	')
                if len(space_split) > 0:
                    result.append(space_split[0])

    return result





from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.Exploration.Visualization.Analysis_Plots import *

class stability_tab(TabBase):

    def __init__(self, parameter, title='Autocorrelation', timesteps=500):
        super().__init__(title)
        self.parameter=parameter
        self.timesteps=timesteps

    def add_recorder_variables(self, neuron_group, Network_UI):
        try:#if hasattr(neuron_group, self.parameter):
            Network_UI.add_recording_variable(neuron_group, 'n.'+self.parameter, timesteps=self.timesteps)
        except:
            print(self.parameter, 'cannot be added to recorder')

    def initialize(self, Network_UI):
        self.stab_tab = Network_UI.Next_Tab(self.title)

        self.image_items = []

        for group_tag1 in Network_UI.neuron_visible_groups:
            self.image_items.append([])
            for group_tag2 in Network_UI.neuron_visible_groups:
                tooltip_msg='x-axis:network average '+self.parameter+' at timestep t\r\ny-axis:corresponding network average '+self.parameter+' at tiemstp t+1'
                self.image_items[-1].append(Network_UI.Add_Image_Item(False, False, title=group_tag1 + self.parameter +'(t) vs ' + group_tag2 + self.parameter +'(t+1)', tooltip_message=tooltip_msg))
            Network_UI.Next_H_Block()

        self.resolution_slider = QSlider(1)  # QtCore.Horizontal
        self.resolution_slider.setMinimum(2)
        self.resolution_slider.setMaximum(200)
        self.resolution_slider.setSliderPosition(50)
        self.resolution_slider.mouseReleaseEvent = Network_UI.static_update_func
        self.resolution_slider.setToolTip('drag to change resolution')
        Network_UI.Add_element(self.resolution_slider)  # , stretch=0.1

    def update(self, Network_UI):
        if self.stab_tab.isVisible():

            for y, group_tag1 in enumerate(Network_UI.neuron_visible_groups):
                for x, group_tag2 in enumerate(Network_UI.neuron_visible_groups):

                    if len(Network_UI.network[group_tag1]) >= 0 and len(Network_UI.network[group_tag2]) >= 0:
                        group1 = Network_UI.network[group_tag1, 0]
                        group2 = Network_UI.network[group_tag2, 0]

                        try:#if hasattr(group1, self.parameter) and hasattr(group2, self.parameter):
                            act1 = np.mean(np.array(group1['n.'+self.parameter, 0][-self.timesteps:]), axis=1)
                            act2 = np.mean(np.array(group2['n.'+self.parameter, 0][-self.timesteps:]), axis=1)
                            self.image_items[y][x].setImage(get_t_vs_tp1_mat(act1, act2, self.resolution_slider.sliderPosition(), False))
                        except:
                            print(self.parameter, 'cannot be evaluated')



from PymoNNto.Exploration.Network_UI.TabBase import *
from PymoNNto.NetworkBehaviour.Recorder.Recorder import *

#from Testing.Common.Grammar_Helper import *

class sidebar_grammar_module_new(TabBase):

    def __init__(self, text_length=30):
        self.text_length = text_length

    def initialize(self, Network_UI):
        if Network_UI.network['STDP', 0] is not None:
            def learning_on_off(event):
                Network_UI.network.set_mechanisms(['STDP'], self.stdp_cb.isChecked())

            self.stdp_cb = QCheckBox()
            self.stdp_cb.setText('STDP')
            self.stdp_cb.setChecked(Network_UI.network['STDP',0].behaviour_enabled)
            self.stdp_cb.stateChanged.connect(learning_on_off)
            Network_UI.Add_Sidebar_Element(self.stdp_cb)

        if Network_UI.network['Text_Activator', 0] is not None:
            def activator_on_off(event):
                Network_UI.network['Text_Activator', 0].behaviour_enabled = self.act_cb.isChecked()

            self.act_cb = QCheckBox()
            self.act_cb.setText('Text input:')
            self.act_cb.setChecked(Network_UI.network['Text_Activator',0].behaviour_enabled)
            self.act_cb.stateChanged.connect(activator_on_off)
            Network_UI.Add_Sidebar_Element(self.act_cb)

            self.inp_text_label = QLabel(Network_UI.main_window)
            Network_UI.Add_Sidebar_Element(self.inp_text_label, stretch=0.2)
            self.inp_text_label.setText('')
            self.inp_text_label.setFont(QFont("Courier"))
            self.inp_text_label.setToolTip('current network input')

        if Network_UI.network['Text_Reconstructor', 0] is not None:
            rl = QLabel(Network_UI.main_window)
            Network_UI.Add_Sidebar_Element(rl, stretch=0.2)
            rl.setText('Reconstruction:')

            self.recon_text_label = QLabel(Network_UI.main_window)
            Network_UI.Add_Sidebar_Element(self.recon_text_label, stretch=0.2)
            self.recon_text_label.setText('')
            self.recon_text_label.setFont(QFont("Courier"))
            self.recon_text_label.setToolTip('simple reconstruction text')

    def update(self, Network_UI):

        if Network_UI.network['Text_Generator', 0] is not None and Network_UI.network['Text_Activator', 0] is not None:
            if Network_UI.network['Text_Activator', 0].behaviour_enabled:
                inp_text = Network_UI.network['Text_Generator', 0].history[-self.text_length:]
            else:
                inp_text = 'deactivated'
            self.inp_text_label.setText(inp_text)

        if Network_UI.network['Text_Reconstructor', 0] is not None:
            recon_text = Network_UI.network['Text_Reconstructor', 0].reconstruction_history[-self.text_length:]
            self.recon_text_label.setText(recon_text)
'''

class sidebar_grammar_module(TabBase):

    def __init__(self, next_p=True, simu_p=True, noc_p=True, text_length=35):
        self.next_p=next_p
        self.simu_p=simu_p
        self.noc_p=noc_p
        self.text_length=text_length


    def add_recorder_variables(self, neuron_group, Network_UI):
        return
        #if hasattr(neuron_group, 'output'):
        #    Network_UI.add_recording_variable(neuron_group, 'n.output', timesteps=100)

    def initialize(self, Network_UI):
        if Network_UI.network['grammar_act', 0] is not None or Network_UI.network['Text_Activator', 0] is not None:

            self.grammar_tab = Network_UI.Next_Tab('text analysis')

            self.readout = None
            self.readout_simu = None

            # self.Add_Sidebar_Spacing()

            def learning_on_off(event):
                Network_UI.network.set_mechanisms(['STDP'], self.stdp_cb.isChecked())

            self.stdp_cb = QCheckBox()
            self.stdp_cb.setText('STDP')
            self.stdp_cb.setChecked(True)
            self.stdp_cb.stateChanged.connect(learning_on_off)
            Network_UI.Add_Sidebar_Element(self.stdp_cb)

            def grammar_activator_on_off(event):
                if Network_UI.network['grammar_act', 0] is not None:
                    Network_UI.network['grammar_act', 0].behaviour_enabled = self.input_select_box.currentText() != 'None'
                if Network_UI.network['Text_Activator', 0] is not None:
                    Network_UI.network['Text_Activator', 0].behaviour_enabled = self.input_select_box.currentText() != 'None'

            self.input_select_box = QComboBox()
            self.input_select_box.addItem("Grammar Act.")
            self.input_select_box.addItem("Prediction")
            self.input_select_box.addItem("None")
            self.input_select_box.currentIndexChanged.connect(grammar_activator_on_off)
            self.input_select_box.setToolTip('select which input is fed into the network')
            Network_UI.Add_Sidebar_Element(self.input_select_box)

            self.inp_text_label = QLabel(Network_UI.main_window)
            Network_UI.Add_Sidebar_Element(self.inp_text_label, stretch=0.2)
            self.inp_text_label.setText('')
            self.inp_text_label.setToolTip('current network input')
            self.text = []

            def train_click(event):

                item = pg.InfiniteLine(pos=Network_UI.network.iteration, movable=False, angle=90)
                self.wnr_plot.addItem(item)

                Network_UI.network.add_behaviours_to_neuron_groups({100: Recorder(['n.output'], tag='pediction_rec')}, Network_UI.network['prediction_source'])
                Network_UI.network.add_behaviours_to_neuron_groups({101: Recorder(['n.pattern_index'], tag='index_rec')}, Network_UI.network['text_input_group'])

                Network_UI.network['grammar_act', 0].behaviour_enabled = True

                steps = 5000
                Network_UI.network.simulate_iterations(steps, 100, measure_block_time=True)

                if self.next_p:
                    self.readout = train(Network_UI.network['pediction_rec'], 'n.output', Network_UI.network['index_rec', 0], 'n.pattern_index', 0, steps, lag=1)

                if self.simu_p:
                    self.readout_simu = train_same_step(Network_UI.network['pediction_rec'], 'n.output', Network_UI.network['index_rec', 0], 'n.pattern_index', 0, steps)


                Network_UI.network.remove_behaviours_from_neuron_groups(Network_UI.network['prediction_source'], tags=['pediction_rec'])
                Network_UI.network.remove_behaviours_from_neuron_groups(Network_UI.network['text_input_group'], tags=['index_rec'])


                self.input_select_box.setCurrentIndex(1)

                print('training_finished')

                item = pg.InfiniteLine(pos=Network_UI.network.iteration, movable=False, angle=90)
                self.wnr_plot.addItem(item)

                item = pg.InfiniteLine(pos=Network_UI.network.iteration+1000, movable=False, angle=90)
                self.wnr_plot.addItem(item)

            if self.noc_p:
                self.noc_text_label = QLabel(Network_UI.main_window)
                Network_UI.Add_Sidebar_Element(self.noc_text_label, stretch=0.2)
                self.noc_text = list(self.noc_text_label.text())
                self.noc_text_label.setToolTip('most active input neuron character')


            if self.next_p:
                self.pred_text_label = QLabel(Network_UI.main_window)
                Network_UI.Add_Sidebar_Element(self.pred_text_label, stretch=0.2)
                self.pred_text_label.mousePressEvent = train_click
                self.pred_text_label.setText('Click to Train...')
                self.pred_text = list(self.pred_text_label.text())
                self.pred_text_label.setToolTip('classifiers prediction for next timesteps (can be fed back into the network)')

            if self.simu_p:
                self.pred_simu_text_label = QLabel(Network_UI.main_window)
                Network_UI.Add_Sidebar_Element(self.pred_simu_text_label, stretch=0.2)
                self.pred_simu_text_label.mousePressEvent = train_click
                self.pred_simu_text_label.setText('Click to Train...')
                self.pred_simu_text = list(self.pred_simu_text_label.text())
                self.pred_simu_text_label.setToolTip('classifiers prediction for current timestep')

            self.text_box = QTextEdit()

            #self.highlighter = Highlighter(self.editor.document())
            #self.editor.setText(sample)
            #layout = QtGui.QVBoxLayout(self)
            #layout.addWidget(self.editor)

            Network_UI.Add_element(self.text_box)
            self.noc_text_long = ''


            self.wnr_tab = Network_UI.Next_Tab('WNR')

            btn = QPushButton('train')
            btn.mousePressEvent = train_click
            Network_UI.Add_element(btn)

            _, self.wnr_bar_plot = Network_UI.Add_plot_curve('Wrong/New/Right Sentences', True, False, legend=False,x_label='', y_label='')

            Network_UI.Next_H_Block()

            self.wrong_s = []
            self.new_s = []
            self.right_s = []
            self.iteration_list = []
            self.wnr_curves, self.wnr_plot = Network_UI.Add_plot_curve('Wrong/New/Right Sentences', True, False, number_of_curves=3, legend=False, x_label='', y_label='', colors=['r', (255,150,0), 'g'])



    def update(self, Network_UI):

        # save data timestep
        if not Network_UI.update_without_state_change and Network_UI.network['grammar_act', 0] is not None:

            grammar_act = Network_UI.network['grammar_act', 0]

            if grammar_act is not None:
                self.inp_text_label.setText('I: ' + ''.join(self.text[-self.text_length:]))

                self.noc_text_label.setText('NC: ' + ''.join(self.noc_text[-self.text_length:]))

                char_act = np.zeros(len(grammar_act.alphabet))
                for ng in Network_UI.network['prediction_source']:
                    recon = ng.Input_Weights.transpose().dot(ng.output)
                    char_act += recon
                char = grammar_act.index_to_char(np.argmax(char_act))

                self.noc_text.append(char)
                self.noc_text_long += char
                if self.grammar_tab.isVisible():
                    self.text_box.setText(grammar_act.mark_with_grammar(self.noc_text_long, False))

                if self.readout_simu is not None:
                    symbol_simu = predict_char(self.readout_simu, Network_UI.network['prediction_source'], 'n.output')
                    char = grammar_act.index_to_char(symbol_simu)
                    self.pred_simu_text += char
                    self.pred_simu_text_label.setText('PC: ' + ''.join(self.pred_simu_text[-self.text_length:]))

                if self.readout is not None:
                    symbol = predict_char(self.readout, Network_UI.network['prediction_source'], 'n.output')
                    char = grammar_act.index_to_char(symbol)
                    self.pred_text += char
                    self.pred_text_label.setText('PN: ' + ''.join(self.pred_text[-self.text_length:]))

                if self.input_select_box.currentText() == 'Prediction':
                    if self.readout is not None:
                        grammar_act.set_next_char(char)
                        # self.network[self.exc_group_name, ts_group].input += self.network['grammar_act', 0].W[:, symbol]sdfdsfgdsfgdf
                    else:
                        print('warning: predictor not trained')

                if self.input_select_box.currentText() == 'Grammar Act.':
                    self.text.append(grammar_act.get_char())
                elif self.input_select_box.currentText() == 'Prediction' and self.readout is not None:
                    self.text.append(char)
                else:
                    self.text.append('|')


            if Network_UI.network.iteration%10 and self.wnr_tab.isVisible():
                score_info = grammar_act.get_text_score(''.join(self.pred_text))

                n_sentences = score_info['n_output_sentences']
                if n_sentences>0:
                    w = score_info['n_wrong_sentences']/n_sentences
                    n = score_info['n_new_sentences']/n_sentences
                    r = score_info['n_right_sentences']/n_sentences

                    self.wrong_s.append(w)
                    self.new_s.append(n)
                    self.right_s.append(r)
                    self.iteration_list.append(Network_UI.network.iteration)

                    self.wnr_curves[0].setData(self.iteration_list, self.wrong_s)
                    self.wnr_curves[1].setData(self.iteration_list, self.new_s)
                    self.wnr_curves[2].setData(self.iteration_list, self.right_s)

                    self.wnr_bar_plot.clear()
                    self.wnr_bar_item = pg.BarGraphItem(x=[1, 2, 3], height=[w, n, r], width=0.6, brushes=['r', (255,150,0), 'g'])
                    self.wnr_bar_plot.addItem(self.wnr_bar_item)

                #self.isi_plt.clear()
                #y, x = np.histogram(SpikeTrain_ISI(self.neuron_act_data), bins=15)
                #curve = pg.PlotCurveItem(x, y, stepMode=True, fillLevel=0, brush=Network_UI.neuron_select_color)
                #self.isi_plt.addItem(curve)

                #if len(self.wrong_s) > 1000: self.wrong_s.pop(0)
                #if len(self.new_s) > 1000: self.new_s.pop(0)
                #if len(self.right_s) > 1000: self.right_s.pop(0)
                #if len(self.iteration_list) > 1000: self.iteration_list.pop(0)

            if len(self.text) > 1000: self.text.pop(0)
            if len(self.pred_text) > 1000: self.pred_text.pop(0)
            if len(self.pred_simu_text) > 1000: self.pred_simu_text.pop(0)


'''
